/*
FILE:     ISTable.h
*/
/*
VERSION:  3.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/* 
  PURPOSE:    Class for indexed string table
*/


#ifndef ISTABLE_H
#define ISTABLE_H


#include <float.h>

#include <string>
#include <vector>
#include <map>

#include "GenString.h"
#include "STable.h"
#include "FileNavigator.h"


using namespace std;


typedef multimap<string, unsigned int, StringCompare> tIndex;


class ISTable : public STable
{
  public:
    enum eSearchType
    {
        eEQUAL = 0,
        eLESS_THAN,
        eLESS_THAN_OR_EQUAL,
        eGREATER_THAN,
        eGREATER_THAN_OR_EQUAL
    };

    enum eSearchDir
    {
        eFORWARD = 0,
        eBACKWARD
    };

  protected:
    // number of digit DBL_MIN_10_EXP, letter e is not included in size
    static const unsigned int EXPONENT      =  4;
    static const unsigned int MAX_PRECISION = DBL_DIG;
    //???DBL_MANT_DIG;
    static const unsigned int MANTISSA       =  MAX_PRECISION + 2;
    static const unsigned int INT_LIMIT      = 11;

    static const unsigned char DT_STRING_VAL = 1; 
    static const unsigned char DT_INTEGER_VAL = 2;
    static const unsigned char DT_DOUBLE_VAL = 3;
    // datatype mask
    static const unsigned char DT_MASK        = 15 << 4;
    // string comparison sensitivity mask
    static const unsigned char SC_MASK        = 0x01;
    // white space sensitivity mask
    static const unsigned char WS_MASK        = 0x02;
    static const unsigned char LAST_DT_VALUE  = 3;
    static const unsigned int  DEFAULT_PRECISION = MAX_PRECISION;
    static const unsigned char DEFAULT_OPTIONS = DT_STRING_VAL << 4;
 
    unsigned int _version;

    vector<unsigned int> _rowMap; 
    unsigned int _numDels;

    vector<unsigned int> _precision;
    vector<char> _compare_opts;

    vector<string> _indexNames;
    vector<vector<unsigned int> > _listsOfColumns;
    vector<tIndex> _indices;
    vector<unsigned int> _unique;

    void Init();
    void Clear();

    inline unsigned int IntRowIndex(const unsigned int rowIndex);

    inline void ClearRowMap();
    void EnlargeRowMap(const unsigned int numRows);
    void ReduceRowMap(const unsigned int numRows);
    void MarkRowDeleted(const unsigned int rowIndex);
    void UnMarkRowDeleted(const unsigned int rowIndex);

    bool IsDelete(const unsigned int rowIndex);

    StringCompare::eCompareType
      GetCompareType(const vector<unsigned int>& colIndices);

    string CellValue(const unsigned int colIndex,
      const unsigned int rowIndex);
    string ConvertString(const string& value, const unsigned int colIndex);
    string MultiStringsValue(const vector<string>& values,
      const vector<unsigned int>& colIndices);
    string SubRowValue(const vector<unsigned int>& colIndices,
      const unsigned int rowIndex);
    string AggregateRow(const vector<unsigned int>& colIndices,
      const unsigned int rowIndex);

    inline void AppendToAndDelimit(string& to, const string& appending);

    void ValidateOptions(unsigned int colIndex);

    string CreateInternalIndexName(const unsigned int indexIndex);
    int UpdateIndex(const unsigned int indexIndex,
      const unsigned int rowIndex);
    int RebuildIndex(const unsigned int indexIndex);
    void ClearIndex(const unsigned int indexIndex);
    void DeleteIndex(const unsigned int indexIndex);

    int FindIndex(const string& indexName);
    int FindIndex(const vector<unsigned int>& colIndices);

    void UpdateIndices(const unsigned int rowIndex);
    void ClearIndices();

    bool IsColumnInIndex(const unsigned int indexIndex,
      const unsigned int colIndex);

    int FindKeyIndex();

    void UpdateColListOnColInsert(const unsigned int colIndex);
    void UpdateColListOnColDelete(const unsigned int colIndex);
    void UpdateColListOnRowAppend(const unsigned int rowIndex,
      const unsigned int fromColIndex, const unsigned int toColIndex);
    void UpdateColListOnCellUpdate(const unsigned int rowIndex,
      const unsigned int colIndex);

    int FindFirst(const unsigned int indexIndex,
      const vector<string>& targets, const vector<unsigned int>& colIndices,
      int& errCode);
    int FindFirst(const unsigned int indexIndex, const vector<string>& targets,
      const vector<string>& colNames, int& errCode);

    void Search(const unsigned int indexIndex, const vector<string>& targets,
      const vector<unsigned int>& colIndices, vector<unsigned int>& res,
      int& errCode, const eSearchType searchType = eEQUAL);
    void Search(const unsigned int indexIndex, const vector<string>& targets,
      const vector<string>& colNames, vector<unsigned int>& res,
      int& errCode, const eSearchType searchType = eEQUAL);

    int GetObjectV3(Word index, FileNavigator *);
    int GetObjectV2(Word index, FileNavigator *);
    int GetObjectV1(Word index, FileNavigator *);
    int GetObjectV1_1(Word index, FileNavigator *);

#ifdef VLAD_DECIDE
    void CompressTable();
#endif
    void MakeTableRectangular();

    void PrintTable(unsigned int indexIndex);

    void ConvertToInt(const string& a, string& ret);
    void ConvertDouble(const string& a, string& ret);
    void ConvertToLowerNoWhiteSpace(const string& a, string& ret);

  public:
    // Sets string comparison case sensitive
    static const char CASE_SENSE = 0x00;
    // Sets string comparison case insensitive
    static const char CASE_INSENSE = 0x01;
    // Sets string comparison to be sensitive to whitespace
    static const char W_SPACE_SENSE = 0x00;
    // Sets string comparison to ignore repeating whitspace.  
    // Also ignores leading and trailing whitespace
    static const char W_SPACE_INSENSE = 0x02;
    // string datatype
    static const char DT_STRING  = DT_STRING_VAL  << 4;
    // integer datatype
    static const char DT_INTEGER = DT_INTEGER_VAL << 4;
    // VLAD FEATURE NOT WORKING double is not working, maybe integer. check it      // double datatype
    static const char DT_DOUBLE  = DT_DOUBLE_VAL  << 4;

    ISTable(const StringCompare::eCompareType colCaseSense =
      StringCompare::eCASE_SENSITIVE) : STable(colCaseSense) { Init(); };

    ISTable(const string& name,
      const StringCompare::eCompareType colCaseSense =
      StringCompare::eCASE_SENSITIVE);

    ~ISTable() { Clear(); };
 
    void Copy(const ISTable& inTable);

    int AddColumn(const string& colName, const char opts = DEFAULT_OPTIONS);
    int InsertColumn(const string& colName, const unsigned int colIndex,
      const char opts = DEFAULT_OPTIONS);
    int InsertColumn(const vector<string>& col, const string& colName,
      const unsigned int colIndex, const char opts = DEFAULT_OPTIONS);
    int FillColumn(const vector<string>& col, const unsigned int colIndex);
    int AppendToColumn(const vector<string>& col, const string& colName);
    int AppendToColumn(const vector<string>& col, const unsigned int colIndex);
    int AddCellToColumn(const string& cell, const unsigned int colIndex);
    int ClearColumn(const unsigned int colIndex);
    int DeleteColumn(const unsigned int colIndex);

    // VLAD: SIMPLIFY: THE CODE DOES NOT COMPILE IF THE DECLARATION AND
    // DEFINITION OF THE SIMPLE GetColumn() IS REMOVED FROM ISTable. I DO NOT
    // KNOW WHY, SINCE THERE IS A METHOD WITH THE SAME NAME IN STable.
    // INVESTIGATE THIS
    void GetColumn(vector<string>& col, const unsigned int colIndex);
    void GetColumn(vector<string>& col, const unsigned int colIndex,
      const string& indexName);

    // VLAD: ERROR HANDLING. AFTER EXCEPTIONS, CHANGE RETURN ARGUMENT TYPE
    int ColumnLength(const unsigned int colIndex);
 
    int AddRow();
    int InsertRow(const unsigned int rowIndex);
    int InsertRow(const vector<string>& row, const unsigned int rowIndex);
    int FillRow(const vector<string>& row, const unsigned int rowIndex);
    int AppendToRow(const vector<string>& row, const unsigned int rowIndex);
    int AddCellToRow(const string& cell, const unsigned int rowIndex);
    int ClearRow(const unsigned int rowIndex);
    int DeleteRow(const unsigned int rowIndex);

    void DeleteRows(const vector<unsigned int>& rows);

    inline unsigned int GetNumRows();
    inline unsigned int GetLastRowIndex();

    void GetRow(vector<string>& row, const unsigned int rowIndex);
    void GetSubRow(vector<string>& subRow, const unsigned int rowIndex,
      const unsigned int fromColIndex, unsigned int toColIndex);

    void FindRedundantRows(const vector<string>& colNames,
      vector<pair<unsigned int, unsigned int> >& duplRows, const int keep,
      const eSearchDir searchDir = eFORWARD);
    void FindRedundantRows(const vector<unsigned int>& colIndices,
      vector<pair<unsigned int, unsigned int> >& duplRows, const int keep,
      const eSearchDir searchDir = eFORWARD);

    int UpdateCell(const string& cell, const unsigned int colIndex,
      const unsigned int rowIndex);

    int GetCell(string& cell, const unsigned int colIndex,
      const unsigned int rowIndex);
    // VLAD: THINK OF HAVING JUST A PURE REFERENCE TO THE OBJECT IN THE TABLE
    // THAT CANNOT BE CHANGED. THIS IS TO AVOID COPYING OF STRINGS AND RETURNING
    // A COPY
    // See how can this be implemented withut pointers but with some
    // references or so
    const string& GetCell(const unsigned int colIndex,
      const unsigned int rowIndex);

    int CreateIndex(const string& indexName, const vector<string>& colNames,
      const int unique = 0);
    int CreateIndex(const string& indexName,
      const vector<unsigned int>& colIndices, const int unique = 0);
    int UpdateIndex(const string& indexName, const unsigned int rowIndex);
    int RebuildIndex(const string& indexName);
    int DeleteIndex(const string& indexName);
    inline unsigned int GetNumIndices();

    inline bool IndexExists(const string& indexName);

    void RebuildIndices();

    int CreateKey(const vector<string>& colNames);
    int CreateKey(const vector<unsigned int>& colIndices);

    int SetFlags(const char newOpts, const unsigned int colIndex);
    int GetDataType(const unsigned int colIndex);

    void SearchColumn(const string& target, const string& colName,
      vector<unsigned int>& res, int& errCode);
    void SearchColumn(const string& target, const unsigned int colIndex,
      vector<unsigned int>& res, int& errCode);

    int FindFirst(const vector<string>& targets, 
      const vector<string>& colNames, int& errCode);
    int FindFirst(const vector<string>& targets, 
      const vector<unsigned int>& colIndices, int& errCode);
    int FindFirst(const string& indexName, const vector<string>& targets,
      const vector<string>& colNames, int& errCode);
    int FindFirst(const string& indexName, const vector<string>& targets, 
      const vector<unsigned int>& colIndices, int& errCode);

    void Search(const vector<string>& targets,
      const vector<string>& colNames, vector<unsigned int>& res, int& errCode,
      const eSearchType searchType = eEQUAL);
    void Search(const vector<string>& targets,
      const vector<unsigned int>& colIndices, vector<unsigned int>& res,
      int& errCode, const eSearchType searchType = eEQUAL);
    void Search(const string& indexName, const vector<string>& targets,
      const vector<unsigned int>& colIndices, vector<unsigned int>& res,
      int& errCode, const eSearchType searchType = eEQUAL);
    void Search(const string& indexName, const vector<string>& targets,
      const vector<string>& colNames, vector<unsigned int>& res,
      int& errCode, const eSearchType searchType = eEQUAL);

    int WriteObject(FileNavigator *);
    int GetObject(Word index, FileNavigator *);

    int Diff(ISTable& sst);

    // TypeOfMerge is 0 for overwrite, 1 for overlap
    int Merge(ISTable& sst, int typeOfMerge = 0); 

    static const string GetErrorMessage(const int errCode);

    void PrintTable();
    void PrintTable(const string& indexName);

    // FOR DEBUGGING ONLY, asserts certain (limited) aspects of internal
    // consistency. More complicated assertions should be constructed to ensure
    // that redundancy lists and trees are correct
    int Assert(unsigned int);

};


inline ISTable::ISTable(const string& name,
  const StringCompare::eCompareType colCaseSense) : STable(name, colCaseSense)
{

    Init();

}


inline unsigned int ISTable::GetNumRows()
{

    return(_numRows - _numDels);

}


inline unsigned int ISTable::GetLastRowIndex()
{

    return(GetNumRows() - 1);

}


inline unsigned int ISTable::GetNumIndices()
{

    return(_indexNames.size());

}


inline bool ISTable::IndexExists(const string& indexName)
{

    int ret = FindIndex(indexName);

    if (ret == -1)
    {
        return(false);
    }
    else
    {
        return(true);
    }

}


inline unsigned int ISTable::IntRowIndex(const unsigned int rowIndex)
{

    // Returns the STable internal row index
    return(_rowMap[rowIndex]);

}


inline void ISTable::ClearRowMap()
{

    _rowMap.clear(); 

}


inline void ISTable::AppendToAndDelimit(string& to, const string& appending)
{

    to += appending;
    // VLAD HARDCODED CONST
    to += " ";

}


#endif // ISTABLE_H
