/*
FILE:     ISTable.C
*/
/*
VERSION:  3.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/* 
  PURPOSE:    Class for indexed string table
*/

#include <sstream>
#include <iostream>
#include <iomanip>

#include "GenString.h"
#include "ISTable.h"


void ISTable::Copy(const ISTable& inTable)
{

    STable::Copy(inTable);

    _version = inTable._version;
    _rowMap = inTable._rowMap;
    _numDels = inTable._numDels;
    _indexNames = inTable._indexNames;
    _listsOfColumns = inTable._listsOfColumns;
    _indices = inTable._indices;
    _unique = inTable._unique;
    _precision = inTable._precision;
    _compare_opts = inTable._compare_opts;

}


int ISTable::AddColumn(const string& colName, const char opts)
{

    return(InsertColumn(colName, _colNames.size(), opts));

}


int ISTable::InsertColumn(const string& colName,
  const unsigned int colIndex, const char opts)
{

    int ret = STable::InsertColumn(colName, colIndex);
    if (ret < 0)
        return ret;

    // VLAD. It is better to introduce one data structure to
    // account for all column info. E.g. we could combine _compare_opts
    // and _precision to be two fields of a struct.
    _compare_opts.insert(_compare_opts.begin() + colIndex, opts);

     // From ANSI C++ standard point of view, the re-casting of
     // DEFAULT_PRECISION to (unsigned int) is not necessary, since its type
     // is unsigned int. However, the linker issues unresolved symbol error
     // if this re-casting is omitted.
    _precision.insert(_precision.begin() + colIndex,
      (unsigned int)DEFAULT_PRECISION);

    SetFlags(opts, colIndex);

    if (colIndex != (_colNames.size() - 1))
    {
        // If it is an insert, the column indices have to be updated.
        UpdateColListOnColInsert(colIndex);
    }

    return(ret);

}


int ISTable::InsertColumn(const vector<string>& col, const string& colName,
  const unsigned int colIndex, const char opts)
{

    InsertColumn(colName, colIndex, opts);

    FillColumn(col, colIndex);

    return(colIndex);

}


int ISTable::FillColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    unsigned int oldNumRows = _numRows;

    int ret;
    if ((ret = STable::FillColumn(col, colIndex)) < NO_TABLE_ERROR)
        return(ret);

    // Clears all indices build on the column colIndex 
    // VLAD: Why do we clear them. Why not rebuilding them?
    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        if (IsColumnInIndex(i, colIndex))
            _indices[i].clear();
    }

    if (_numRows > oldNumRows)
    {
        // If filling of column created new rows, add the spaceholders
        // for them in the row map.
        EnlargeRowMap(_numRows - oldNumRows);
    }

    return(NO_TABLE_ERROR);

}


int ISTable::AppendToColumn(const vector<string>& col, const string& colName)
{

    // VLAD: INDEX CORRUPTION
    // This is not safe after index/indices is/are created
    // indices will be corrupted

    int colIndex = GetColumnIndex(colName);
    if (colIndex < 0)
        return(colIndex);

    return(AppendToColumn(col, colIndex));

}


int ISTable::AppendToColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    // VLAD: INDEX CORRUPTION
    // This is not safe after index/indices is/are created
    // indices will be corrupted

    if (colIndex >= _colNames.size())
        return(COLUMN_OUT_OF_BOUNDS);
    if (col.empty())
        return(ADD_UPDATE_NULL);

    unsigned int oldNumRows = _numRows;

    int ret;
    if ((ret = STable::AppendToColumn(col, colIndex)) < NO_TABLE_ERROR)
        return(ret);

    if (_numRows > oldNumRows)
    {
        // If appending to column created new rows, add the spaceholders
        // for them in the row map.
        EnlargeRowMap(_numRows - oldNumRows);
    }

    return(NO_TABLE_ERROR);

}


int ISTable::AddCellToColumn(const string& cell, const unsigned int colIndex)
{

    // VLAD: INDEX CORRUPTION
    // This is not safe after index/indices is/are created
    // indices will be corrupted

    if (colIndex >= _colNames.size())
        return(COLUMN_OUT_OF_BOUNDS);
    if (cell.empty())
        return(ADD_UPDATE_NULL);

    unsigned int oldNumRows = _numRows;

    int ret;
    if ((ret = STable::AddCellToColumn(cell, colIndex)) < NO_TABLE_ERROR)
        return(ret);

    if (_numRows > oldNumRows)
    {
        // If appending to column created new rows, add the spaceholders
        // for them in the row map.
        EnlargeRowMap(_numRows - oldNumRows);
    }

    return(NO_TABLE_ERROR);

}


int ISTable::ClearColumn(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return(COLUMN_OUT_OF_BOUNDS);

    unsigned int oldNumRows = _numRows;

    int ret;
    if ((ret = STable::ClearColumn(colIndex)) < NO_TABLE_ERROR)
        return(ret);

    if (oldNumRows > _numRows)
    {
        // If clearing of column deleted some rows at the end, remove them from
        // row map.
        ReduceRowMap(oldNumRows - _numRows);
    }

    // Clear all indices built on the column colIndex
    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        if (IsColumnInIndex(i, colIndex))
            _indices[i].clear();
    }

    return(NO_TABLE_ERROR);

}


int ISTable::DeleteColumn(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    unsigned int oldNumRows = _numRows;

    STable::DeleteColumn(colIndex);

    _precision.erase(_precision.begin() + colIndex);
    _compare_opts.erase(_compare_opts.begin() + colIndex);

    if (oldNumRows > _numRows)
    {
        // If deleting of column deleted some rows at the end, remove them
        // from row map.
        ReduceRowMap(oldNumRows - _numRows);
    }

    // Deletes all indices build on the column colIndex 
    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        if (IsColumnInIndex(i, colIndex))
        {
            DeleteIndex(i);
        }
    }

    UpdateColListOnColDelete(colIndex);

    return NO_TABLE_ERROR;

}


// VLAD: SIMPLIFY: THIS SHOULD NOT EXIST, BUT SHOULD BE AUTOMATICALLY INVOKED
void ISTable::GetColumn(vector<string>& col, const unsigned int colIndex)
{
 
    STable::GetColumn(col, colIndex);

}


void ISTable::GetColumn(vector<string>& col, const unsigned int colIndex,
  const string& indexName)
{

    // VLAD: ERROR HANDLING
    if (colIndex >= _colNames.size())
        return;

    unsigned int indexIndex = FindIndex(indexName);

    col.reserve(_indices[indexIndex].size());

    for (tIndex::iterator pos = _indices[indexIndex].begin();
      pos != _indices[indexIndex].end(); ++pos)
    {
        col.push_back((*_data[colIndex])[IntRowIndex(pos->second)]);
    }

}


int ISTable::ColumnLength(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    return(_data[colIndex]->size());

}


int ISTable::AddRow()
{

    return(InsertRow(GetNumRows()));

}


int ISTable::InsertRow(const unsigned int rowIndex)
{

    if (_numDels != 0)
    {
        UnMarkRowDeleted(rowIndex);
    }
    else
    {
        int ret = STable::InsertRow(rowIndex);
        if (ret < NO_TABLE_ERROR)
            return ret;

        EnlargeRowMap(1);
    }

    // VLAD: SEE IF WE NEED TO UPDATE INDICES FOR ADD ROW
    RebuildIndices();

    return(_numRows);

}


int ISTable::InsertRow(const vector<string>& row,
  const unsigned int rowIndex)
{

    int ret = InsertRow(rowIndex);
    if (ret < TABLE_WARNING)
        return ret;

    ret = FillRow(row, rowIndex);
    if (ret < TABLE_WARNING)
        return ret;

    return(_numRows);

}


int ISTable::FillRow(const vector<string>& row, const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
        return ROW_OUT_OF_BOUNDS;

    int ret = STable::FillRow(row, IntRowIndex(rowIndex));
    if (ret < NO_TABLE_ERROR)
        return ret;

    UpdateIndices(rowIndex);

    return NO_TABLE_ERROR;

}


int ISTable::AppendToRow(const vector<string>& row,
  const unsigned int rowIndex)
{

    // VLAD: ERROR HANDLING
    if (rowIndex >= GetNumRows())
        return ROW_OUT_OF_BOUNDS;
    if (row.empty())
        return ADD_UPDATE_NULL;

    unsigned int fromColIndex, toColIndex;

    STable::GetRowAppendRange(row, IntRowIndex(rowIndex), fromColIndex,
      toColIndex);

    STable::FillRow(row, IntRowIndex(rowIndex), fromColIndex, toColIndex);

    UpdateColListOnRowAppend(rowIndex, fromColIndex, toColIndex);

    return NO_TABLE_ERROR;

}


int ISTable::AddCellToRow(const string& cell, const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
        return ROW_OUT_OF_BOUNDS;

    unsigned int colIndex;

    STable::GetCellAddIndex(IntRowIndex(rowIndex), colIndex);

    (*_data[colIndex])[IntRowIndex(rowIndex)] = cell;

    UpdateColListOnCellUpdate(rowIndex, colIndex);

    return NO_TABLE_ERROR;

}



int ISTable::ClearRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
        return ROW_OUT_OF_BOUNDS;

    STable::ClearRow(IntRowIndex(rowIndex));

    UpdateIndices(rowIndex);

    return NO_TABLE_ERROR;

}


int ISTable::DeleteRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows()) return ROW_OUT_OF_BOUNDS;

    STable::ClearRow(IntRowIndex(rowIndex));

    MarkRowDeleted(rowIndex);

    // VLAD - THINK OF MORE EFFICIENT INDEX RE-BUILDING IF POSSIBLE
    RebuildIndices();

    return NO_TABLE_ERROR;

}


void ISTable::DeleteRows(const vector<unsigned int>& rows)
{

    // VLAD ERROR CHECKING ALL THE INDICES in rows
    // VLAD IMPROVE WITH REVERSE ITERATORS or more meaningfull index loops

    for (unsigned int rowIndex = 0; rowIndex < rows.size(); ++rowIndex)
    {
        DeleteRow(rows[rows.size() - 1 - rowIndex]);
    }

}


void ISTable::GetRow(vector<string>& row, const unsigned int rowIndex)
{

    GetSubRow(row, rowIndex, 0, _colNames.size() - 1);

}


void ISTable::GetSubRow(vector<string>& subRow, const unsigned int rowIndex,
  const unsigned int fromColIndex, unsigned int toColIndex)
{

    if (rowIndex >= GetNumRows()) return;
    if (toColIndex < fromColIndex) return;
    if (fromColIndex >= _colNames.size()) return;
    // VLAD: DO WE ALLOW THIS COLUMN INDEX CORRECTION. CHECK ALL THE PLACES
    // with toColIndex
    if (toColIndex >= _colNames.size()) toColIndex = _colNames.size() - 1;

    STable::GetSubRow(subRow, IntRowIndex(rowIndex), fromColIndex, toColIndex);

}


void ISTable::FindRedundantRows(const vector<string>& colNames,
  vector<pair<unsigned int, unsigned int> >& duplRows, const int keep,
  const eSearchDir searchDir)
{

    vector<unsigned int> colIndices;

    int tableIndex;

    // VLAD: MAKE AND UTILIZE METHOD THAT CONVERTS FROM COLUMN NAMES
    // TO COLUMN INDICES
    // VLAD VALUE replace with GetColumnIndices()
    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        tableIndex = GetColumnIndex(colNames[i]);
        if (tableIndex >= 0)
  	    colIndices.push_back(tableIndex);
        else
        {
            // VLAD ERROR HANDLING - THROW EXCEPTION
            ;
        }
    }

    FindRedundantRows(colIndices, duplRows, keep, searchDir);

}


void ISTable::FindRedundantRows(const vector<unsigned int>& colIndices,
  vector<pair<unsigned int, unsigned int> >& duplRows, const int keep,
  const eSearchDir searchDir)
{

    // VLAD ERROR HANDLING - CHECK INDICES
    // for keep = 1, reports redundant rows, but keeps them in table
    // for keep = 0, reports redundant rows and deletes them from table
    // searchDir can be forward or backward
 
    unsigned int realRowIndex;
    tIndex::iterator pos;
    string value;

    duplRows.clear();

    MakeTableRectangular();

    StringCompare::eCompareType compareType = GetCompareType(colIndices);

    tIndex tmpindex(compareType);

    for (unsigned int rowIndex = 0; rowIndex < GetNumRows(); rowIndex++)
    {
        if (searchDir == eFORWARD)
        {
            realRowIndex = rowIndex;
        } 
        else
        {
            realRowIndex = GetNumRows() - rowIndex - 1;
        }

        value = AggregateRow(colIndices, realRowIndex);

        pos = tmpindex.find(value);
        if (pos != tmpindex.end())
        {
            duplRows.push_back(make_pair(realRowIndex, pos->second));
        }
        else
        {
            tIndex::value_type valType(value, realRowIndex);
            tmpindex.insert(valType);
        }

        value.clear();
    }

    if (keep == 0)
    {
        vector<unsigned int> duplRowIndices;
        for (unsigned int rowI = 0; rowI < duplRows.size(); rowI++)
        {
            duplRowIndices.push_back(duplRows[rowI].first);
        }

        DeleteRows(duplRowIndices);
    }

}


int ISTable::UpdateCell(const string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (colIndex >= _colNames.size()) return COLUMN_OUT_OF_BOUNDS;
    if (rowIndex >= GetNumRows()) return ROW_OUT_OF_BOUNDS;

    STable::UpdateCell(cell, colIndex, IntRowIndex(rowIndex));

    UpdateColListOnCellUpdate(rowIndex, colIndex);

    return NO_TABLE_ERROR;

}


int ISTable::GetCell(string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (colIndex >= _colNames.size()) return COLUMN_OUT_OF_BOUNDS;
    // VLAD. Why is here ->size() instead of GetNumRows()?
    if (rowIndex >= _data[colIndex]->size()) return ROW_OUT_OF_BOUNDS;
    cell = (*_data[colIndex])[IntRowIndex(rowIndex)];
    return NO_TABLE_ERROR;

}


const string& ISTable::GetCell(const unsigned int colIndex,
  const unsigned int rowIndex)
{
  // VLAD - ADD - ERROR HANDLING CODE
  // if ((colIndex < 0) || (colIndex >= _colNames.size())) return COLUMN_OUT_OF_BOUNDS;
  // if ((rowIndex < 0) || (rowIndex >= (int)(_data[colIndex]->size()))) return ROW_OUT_OF_BOUNDS;
  return (*(_data[colIndex]))[IntRowIndex(rowIndex)];
  // return NO_TABLE_ERROR;
}


int ISTable::CreateIndex(const string& indexName,
  const vector<string>& colNames, const int unique)
{

    //   if unique flag is set to 0, table rows with duplicate values
    //     are allowed
    //   if unique flag is set to 1, table must not have duplicate values
    //     and rows with repeated values are deleted.

    vector<unsigned int> colIndices;

    int ret = GetColumnsIndices(colNames, colIndices);

    if (ret == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        return(ret);
    }

    return(CreateIndex(indexName, colIndices, unique));

}


int ISTable::CreateIndex(const string& indexName,
  const vector<unsigned int>& colIndices, const int unique)
{

    if (colIndices.empty())
    {
        return(NO_TABLE_ERROR);
    }

    int ret = VerifyColumnsIndices(colIndices);
    if (ret != NO_TABLE_ERROR)
    {
        return(ret);
    }

    MakeTableRectangular() ;

    int indexIndex = FindIndex(indexName);
    if (indexIndex != -1)
    {
        // Index with the same name found. Just update it.
        ClearIndex(indexIndex);
        _listsOfColumns[indexIndex] = colIndices;
        _unique[indexIndex] = unique;
    }
    else
    {
        StringCompare::eCompareType compareType = GetCompareType(colIndices);

        // Create one or another type of multimap with/without
        // case-insensitivity
        tIndex index(compareType);

        _indexNames.push_back(indexName);
        _listsOfColumns.push_back(colIndices);
        _indices.push_back(index);
        _unique.push_back(unique);
    }
 
    for (unsigned int i = 0; i < GetNumRows(); i++)
        UpdateIndex(indexName, i);

    return(NO_TABLE_ERROR);

}


int ISTable::UpdateIndex(const string& indexName, const unsigned int rowIndex)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
        return INDEX_NAME_NOT_FOUND;

    UpdateIndex(indexIndex, rowIndex);

    return NO_TABLE_ERROR;

}


int ISTable::RebuildIndex(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
        return INDEX_NAME_NOT_FOUND;

    RebuildIndex(indexIndex);

    return(NO_TABLE_ERROR);

}


int ISTable::DeleteIndex(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
        return INDEX_NAME_NOT_FOUND;

    DeleteIndex(indexIndex);

    return(NO_TABLE_ERROR);

}


void ISTable::RebuildIndices()
{

    for (unsigned int i = 0; i < _indices.size(); i++)
        RebuildIndex(i);

}


int ISTable::CreateKey(const vector<string>& colNames)
{

    vector<unsigned int> colIndices;

    int ret = GetColumnsIndices(colNames, colIndices);

    if (ret == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        return(ret);
    }

    return(CreateKey(colIndices));

}


int ISTable::CreateKey(const vector<unsigned int>& colIndices)
{

    // VLAD HARDCODED CONST
    return(CreateIndex("__key__", colIndices, 1));

}


int ISTable::SetFlags(const char newOpts, const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return COLUMN_OUT_OF_BOUNDS;

    if (newOpts & DT_MASK)
    {
        _compare_opts[colIndex] &= 15;
        _compare_opts[colIndex] |= ((~0) & ( newOpts & DT_MASK ));
    }

    if (newOpts & CASE_INSENSE)
    {
        _compare_opts[colIndex] |= CASE_INSENSE;
    }
    else
        _compare_opts[colIndex] &= ~CASE_INSENSE;

    if (newOpts & W_SPACE_INSENSE)
        _compare_opts[colIndex] |= W_SPACE_INSENSE;
    else
        _compare_opts[colIndex] &= ~W_SPACE_INSENSE;

    ValidateOptions(colIndex);

    return NO_TABLE_ERROR;

}


int ISTable::GetDataType(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return 0;

    return((_compare_opts[colIndex] & DT_MASK) >> 4);

}


int ISTable::UpdateIndex(const unsigned int indexIndex,
  const unsigned int rowIndex)
{

    // indexIndex is the index of an index
    // rowIndex is the row index
    // This function preforms the following:
    //   if unique flag is set to 0, table rows with duplicate values
    //     are allowed
    //   if unique flag is set to 1, table must not have duplicate values
    //     and rows with repeated values are deleted.
    //
    // More information on unique flag being set to 1. If a row with rowIndex
    // already exists and the value of the new row with rowIndex already
    // exists, delete the row with rowIndex from the table and from the index.
    // VLAD: SIMPLIFY THE ABOVE unique == 1 processing.

    // VLAD ERROR HANDLING
    if (rowIndex >= GetNumRows())
        return ROW_OUT_OF_BOUNDS;

    string value = SubRowValue(_listsOfColumns[indexIndex], rowIndex);

    if (_unique[indexIndex] == 0)
    {
        // Duplicate values are allowed, but not duplicate row indices.

        // VLAD IMPROVE. This and below for loop can be made into a method
        // IndexSearchForRow()

        // See if a pair with the row having rowIndex already exists. If it
        // does delete it first.
        for (tIndex::iterator pos = _indices[indexIndex].begin();
          pos != _indices[indexIndex].end(); ++pos)
        {
            if (pos->second == rowIndex)
            {
                _indices[indexIndex].erase(pos);
                break;
            }
        }

        // Duplicate values are allowed, just store value/index pair.
        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);

        return NO_TABLE_ERROR;
    }

    // Assuming that _unique[indexIndex] is 1
    pair<tIndex::iterator, tIndex::iterator> range;

    range = _indices[indexIndex].equal_range(value);

    if (range.first == range.second)
    {
        // Nothing found. Just store value/index pair.
        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);
    }
    else
    {
        // Found a value. If a row with rowIndex already exists,
        // remove it from the indices and delete it from the table.
        for (tIndex::iterator pos = _indices[indexIndex].begin();
          pos != _indices[indexIndex].end(); ++pos)
        {
            if (pos->second == rowIndex)
            {
                _indices[indexIndex].erase(pos);
                DeleteRow(rowIndex);
                break;
            }
        }
    }

    return NO_TABLE_ERROR;

}


int ISTable::RebuildIndex(const unsigned int indexIndex)
{

    _indices[indexIndex].clear();

    string value;

    for (unsigned int rowIndex = 0; rowIndex < GetNumRows(); rowIndex++)
    {
        value = SubRowValue(_listsOfColumns[indexIndex], rowIndex);

        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);

        value.clear();
    }

    return NO_TABLE_ERROR;

}


void ISTable::ClearIndex(const unsigned int indexIndex)
{

    _listsOfColumns[indexIndex].clear();
    _indices[indexIndex].clear();

}


void ISTable::DeleteIndex(const unsigned int indexIndex)
{

    // VLAD: THIS MAY NOT BE NEEDED. DEPENDS ON HOW STL CONTAINERS
    // FREE RESOURCES IN ERASE CALLS
    ClearIndex(indexIndex);  

    _indexNames.erase(_indexNames.begin() + indexIndex);
    _listsOfColumns.erase(_listsOfColumns.begin() + indexIndex);
    _indices.erase(_indices.begin() + indexIndex);
    _unique.erase(_unique.begin() + indexIndex);

}


int ISTable::FindIndex(const string& indexName)
{

    // VLAD: FOR LOOP SEARCH: Use STL construct given below for this search
    // int index = distance(strings.begin(), find(strings.begin(), 
    // strings.end(), string("xyz"))); 
    for (unsigned int i = 0; i < _indexNames.size(); i++)
    {
        if (_indexNames[i] == indexName)
        {
            return(i);
        }
    }

    return(-1);

}


int ISTable::FindIndex(const vector<unsigned int>& colIndices)
{

  int equal;

  if (colIndices.empty())
  {
      return(-1);
  }

  StringCompare::eCompareType compareType = GetCompareType(colIndices);

  for (unsigned int i = 0; i < _indices.size(); i++)
  {
      if (_listsOfColumns[i].size() != colIndices.size())
      {
          continue;
      }

      // VLAD INDEX SIMPLIFY CompareColIndices()
      // Assume that all columns are equal
      equal = 0;
      for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
      {
          if (_listsOfColumns[i][j] != colIndices[j])
          {
              // Found one non-equal column. Set the flag and brake the loop.
              equal = 1;
              break;
          }
      }

      if (equal == 0)
      {
          if ((_indices[i].key_comp()).GetCompareType() == compareType)
          {
              // The index is matched only if it also has the same
              // compare type
              return(i);
          } 
      }
  }

  return(-1);

}


void ISTable::UpdateIndices(const unsigned int rowIndex)
{

    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        UpdateIndex(i, rowIndex);
    }

}


void ISTable::ClearIndices()
{

    for(unsigned int i = 0; i < _indexNames.size(); i++)
    {
        ClearIndex(i);
    }

}


bool ISTable::IsColumnInIndex(const unsigned int indexIndex,
  const unsigned int colIndex)
{

    vector<unsigned int>::iterator pos =
      find(_listsOfColumns[indexIndex].begin(),
      _listsOfColumns[indexIndex].end(), colIndex);

    if (pos != _listsOfColumns[indexIndex].end())
    {
        return(true);
    }
    else
    {
        return(false);
    }

}


int ISTable::FindKeyIndex()
{

    // VLAD HARDCODED CONST
    return(FindIndex("__key__"));

}


string ISTable::CellValue(const unsigned int colIndex,
  const unsigned int rowIndex)
{

    return(ConvertString((*_data[colIndex])[IntRowIndex(rowIndex)], colIndex));

}


string ISTable::ConvertString(const string& value,
  const unsigned int colIndex)
{
  string ret;
  unsigned char temp = _compare_opts[colIndex];
  switch (( temp & DT_MASK) >> 4) {
  case DT_INTEGER_VAL: {
      ConvertToInt(value, ret);
      break;
  }
  case DT_DOUBLE_VAL: {
      ConvertDouble(value, ret);
      break;
  }
  case DT_STRING_VAL: {
      if ( temp & SC_MASK ) {
        if ( temp & WS_MASK ) {
          ConvertToLowerNoWhiteSpace(value, ret);
	  break;
	}
        else {
          String::LowerCase(value, ret);
	  break;
	}
      } else if (temp & WS_MASK) {
	String::RemoveWhiteSpace(value, ret);
	break;
      } else {
	ret=value;
	break;
      }
  }
  default: {
      ret=value;
      break;
  }
  }
  return ret;
}


string ISTable::MultiStringsValue(const vector<string>& values, 
    const vector<unsigned int>& colIndices)
{

    string value;

    for (unsigned int index = 0; index < values.size(); index++)
    {
        AppendToAndDelimit(value,
          ConvertString(values[index], colIndices[index]));
    }

    return(value);

}


string ISTable::SubRowValue(const vector<unsigned int>& colIndices,
  const unsigned int rowIndex)
{

    string value;

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        AppendToAndDelimit(value, CellValue(colIndices[index], rowIndex));
    }

    return(value);

}


string ISTable::AggregateRow(const vector<unsigned int>& colIndices,
  const unsigned int rowIndex)
{

    string value;
    string cell;

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        GetCell(cell, colIndices[index], rowIndex);
        AppendToAndDelimit(value, cell);
    }

    return(value);

}


int ISTable::Assert(unsigned int colIndex) {
  if (colIndex >= _colNames.size())
  {
    return COLUMN_OUT_OF_BOUNDS;
  }

  if (!_data[colIndex]) {
    return ASSERT_NULL_DATA_POINTER;
  }

  if (_indexNames.size()!=_listsOfColumns.size()!= _indices.size())
    return INDEX_CORRUPTED;
  
  return NO_TABLE_ERROR;
}


void ISTable::Init()
{

    // VLAD: MAKE THIS NON-HARD-CODED AND SEE ALL OTHER PLACES IN THE CODE
    // FOR VERSION USAGE
    _version = 5;
    _numDels = 0;

}


void ISTable::Clear()
{

    // VLAD: MAKE THIS NON-HARD-CODED AND SEE ALL OTHER PLACES IN THE CODE
    // FOR VERSION USAGE
    STable::Clear();

    _version = 5;
    _rowMap.clear();
    _numDels = 0;
    _indexNames.clear();
    _listsOfColumns.clear();
    _indices.clear();
    _unique.clear();
    _precision.clear();
    _compare_opts.clear();

}


void ISTable::ValidateOptions(unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
        return;

    unsigned char a = 0, b = 0;

    b = _compare_opts[colIndex];

    a = b & DT_MASK;

    if (((a >> 4) > LAST_DT_VALUE) || (a == 0))
    {
        a = DEFAULT_OPTIONS;
    }
    else
        a = b;

    a &= ~(3 << 2); // not necessary, but clean up 3rd and 4th bit

    _compare_opts[colIndex] = a;

}


string ISTable::CreateInternalIndexName(const unsigned int indexIndex)
{

    // VLAD HARCODED CONST
    string name("index_");

    name += String::IntToString(indexIndex);

    return(name);

}


void ISTable::SearchColumn(const string& target, const string& colName,
  vector<unsigned int>& res, int& errCode)
{

    int colIndex = GetColumnIndex(colName);
    if (colIndex < 0)
    {
        return;
    }

    SearchColumn(target, colIndex, res, errCode);

}


void ISTable::SearchColumn(const string& target,
  const unsigned int colIndex, vector<unsigned int>& res, int& errCode)
{

    vector<string> targets;
    targets.push_back(target);

    vector<unsigned int> colIds;

    colIds.push_back(colIndex);

    Search(targets, colIds, res, errCode, eEQUAL);

}


void ISTable::Search(const vector<string>& targets,
  const vector<string>& colNames, vector<unsigned int>& res, int& errCode,
  const eSearchType searchType)
{

    int eCode2 = 0;

    vector<unsigned int> colIds;
    int retCode = GetColumnsIndices(colNames, colIds);

    if (retCode == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    if (colIds.empty())
    {
        // VLAD eCode2 WORK
        if (eCode2 >= TABLE_WARNING && eCode2 < 0)
            errCode = eCode2;
        return;
    }

    Search(targets, colIds, res, errCode, searchType);
    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

}


void ISTable::Search(const vector<string>& targets,
  const vector<unsigned int>& colIds, vector<unsigned int>& res,
  int& errCode, const eSearchType searchType)
{

    int indexIndex = FindIndex(colIds);

    if (indexIndex == -1)
    {
        // No search index found. Create a new one.
        indexIndex = _indices.size();
        string name = CreateInternalIndexName(indexIndex);
        CreateIndex(name, colIds);
    }

    Search(indexIndex, targets, colIds, res, errCode, searchType);

}


void ISTable::Search(const unsigned int indexNum,
  const vector<string>& targets, const vector<unsigned int>& colIds,
  vector<unsigned int>& res, int& errCode, const eSearchType searchType)
{

    errCode = NO_TABLE_ERROR;

    res.clear();

    unsigned int len = targets.size();
    if ((len > colIds.size()) || (len < 1))
    {
        errCode = COLUMN_OUT_OF_BOUNDS;
        return;
    }

    int retCode = VerifyColumnsIndices(colIds);
    if (retCode != NO_TABLE_ERROR)
    {
        errCode = retCode;
        return;
    }

    string value = MultiStringsValue(targets, colIds);

    /*
    ** This code searches the container for all key/value pairs, where key
    ** equals "value" and stores all the values in the ret.
    */

    pair<tIndex::iterator, tIndex::iterator> range;

    range = _indices[indexNum].equal_range(value);

    tIndex::iterator start, afterLast;

    switch (searchType)
    {
        case eEQUAL:
           start = range.first;
           afterLast = range.second;
           break;
        case eLESS_THAN:
           start = _indices[indexNum].begin();
           afterLast = range.first;
           break;
        case eLESS_THAN_OR_EQUAL:
           start = _indices[indexNum].begin();
           afterLast = range.second;
           break;
        case eGREATER_THAN:
           start = range.second;
           afterLast = _indices[indexNum].end();
           break;
        case eGREATER_THAN_OR_EQUAL:
           start = range.first;
           afterLast = _indices[indexNum].end();
           break;
        default:
           // VLAD ERROR CHECKING: DO THIS CHECK PRIOR TO SEARCH
           errCode = COLUMN_OUT_OF_BOUNDS;
           return;
           break; 
    }

    for (tIndex::iterator pos = start; pos != afterLast; ++pos)
    {
        res.push_back(pos->second);
    }

    if (res.empty())
    {
        errCode = NOT_FOUND;
    }

    if (searchType == eEQUAL)
        sort(res.begin(), res.end());

}


void ISTable::Search(const unsigned int indexIndex,
  const vector<string>& targets, const vector<string>& colNames,
  vector<unsigned int>& res, int& errCode, const eSearchType searchType)
{

    int eCode2 = 0;
    vector<unsigned int> colIds;
    int retCode = GetColumnsIndices(colNames, colIds);

    if (retCode == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    Search(indexIndex, targets, colIds, res, errCode, searchType);
    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

}


void ISTable::Search(const string& indexName,
  const vector<string>& targets, const vector<unsigned int>& colIds,
  vector<unsigned int>& res, int& errCode, const eSearchType searchType)
{
 
    errCode = NO_TABLE_ERROR;

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
    {
        errCode = INDEX_NAME_NOT_FOUND;
        return;
    }

    Search(indexIndex, targets, colIds, res, errCode, searchType);

}


void ISTable::Search(const string& indexName,
  const vector<string>& targets, const vector<string>& colNames,
  vector<unsigned int>& res, int& errCode, const eSearchType searchType)
{

    int eCode2 = 0;
    vector<unsigned int> colIds;
    int retCode = GetColumnsIndices(colNames, colIds);

    if (retCode == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    Search(indexName, targets, colIds, res, errCode, searchType);

    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

}


int ISTable::FindFirst(const vector<string>& targets,
  const vector<string>& colNames, int& errCode)
{

    // VLAD: ADD CODE TO CHECK IF THE SIZES OF targets and colNames
    // are different and to return error.

    int eCode2 = 0;
    vector<unsigned int> colIds;
    int ret = GetColumnsIndices(colNames, colIds);

    if (ret == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    ret = FindFirst(targets, colIds, errCode);
    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

    return(ret);

}


int ISTable::FindFirst(const unsigned int indexIndex,
  const vector<string>& targets, const vector<string>& colNames,
  int& errCode)
{

    int eCode2 = 0;
    vector<unsigned int> colIds;
    int ret = GetColumnsIndices(colNames, colIds);

    if (ret == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    ret = FindFirst(indexIndex, targets, colIds, errCode);
    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

    return(ret);

}


int ISTable::FindFirst(const string& indexName,
  const vector<string>& targets, const vector<string>& colNames,
  int& errCode)
{

    int eCode2 = 0;
    vector<unsigned int> colIds;
    int ret = GetColumnsIndices(colNames, colIds);

    if (ret == SOME_COLUMN_NAMES_NOT_FOUND)
    {
        eCode2 = SOME_COLUMN_NAMES_NOT_FOUND;
    }

    ret = FindFirst(indexName, targets, colIds, errCode);
    // VLAD eCode2 WORK
    if (eCode2 >= TABLE_WARNING && eCode2 < 0)
        errCode = eCode2;

    return(ret);

}


int ISTable::FindFirst(const vector<string>& targets,
  const vector<unsigned int>& colIds, int& errCode)
{

    int indexIndex = FindIndex(colIds);
    if (indexIndex == -1)
    {
        indexIndex = _indices.size();
        string name = CreateInternalIndexName(indexIndex);
        CreateIndex(name, colIds);
    }

    return(FindFirst(indexIndex, targets, colIds, errCode));

}


int ISTable::FindFirst(const string& indexName,
  const vector<string>& targets, const vector<unsigned int>& colIds,
  int& errCode)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
        return INDEX_NAME_NOT_FOUND;

    return(FindFirst(indexIndex, targets, colIds, errCode));

}


int ISTable::FindFirst(const unsigned int indexNum,
  const vector<string>& targets, const vector<unsigned int>& colIds,
  int& errCode)
{

    int ret = -1;

    if (colIds.empty())
    {
        errCode = NOT_FOUND;
        return(ret);
    }

    unsigned int len = targets.size();
    if ((len < colIds.size()) || (len < 1))
    {
        errCode = COLUMN_OUT_OF_BOUNDS;
        return(ret);
    }

    int retCode = VerifyColumnsIndices(colIds);
    if (retCode != NO_TABLE_ERROR)
    {
        errCode = retCode;
        return(ret);
    }

    string value = MultiStringsValue(targets, colIds);
 
    // Assume that it is not found
    errCode = NOT_FOUND;
    tIndex::iterator pos = _indices[indexNum].find(value);
    if (pos != _indices[indexNum].end())
    {
        // Found
        ret = pos->second;
        errCode = NO_TABLE_ERROR;
    }

    return(ret);

}


int ISTable::Merge(ISTable& inTable,int typeOfMerge) {
  unsigned int i,j;
  int index;
  vector<string> ColNames1;
  vector<string> ColNames2;
  vector<string> target;
  int errCode;
  string cell;
  string empty("");

  GetColumnNames(ColNames1);
  inTable.GetColumnNames(ColNames2);
  
  
  vector<string> newColumn;
  // VLAD: EXAMINE IF CURRENT MERGE OPERATION WORKS AS IT SHOULD FROM THE
  // PERSPECTIVE OF _numDels
  // VLAD: GetNumRows()?
  for (i=0; i<_numRows; i++)
    newColumn.push_back(empty);
  for (i=0; i<inTable._colNames.size(); i++) {
    if (GetColumnIndex(ColNames2[i])<0) {
      AddColumn(ColNames2[i]);
      FillColumn(newColumn,_colNames.size()-1);
      // VLAD Do we need to EnlargeRowMap()?
    }
  }
  ColNames1.clear();
  GetColumnNames(ColNames1);
  vector<string> newRow;
  // VLAD: GetNumRows()?
  for (i=0; i<inTable._numRows; i++) {
    for (j=0; j<_colNames.size(); j++) {
      cell.clear();
      if ((index = inTable.GetColumnIndex(ColNames1[j]))>=0) {
	inTable.GetCell(cell,index,i);
	newRow.push_back(cell);
      }
      else
	newRow.push_back(empty);
    }
    
    // Find out what column ids are in key index
    int indexIndex = FindKeyIndex();
    for (unsigned int jj=0; jj< inTable._listsOfColumns[indexIndex].size();
      jj++) {
      cell.clear();
      inTable.GetCell(cell,inTable._listsOfColumns[indexIndex][jj],i);
      target.push_back(cell);
    }
    vector<unsigned int> result; 
    Search(indexIndex, target,_listsOfColumns[indexIndex],result, errCode);
    if (result.empty()) {
      AddRow();
      FillRow(newRow,GetLastRowIndex());
    }
    else {
      if (typeOfMerge == 1) {
	// overlap
	for (j=0; j<_colNames.size(); j++) {
	  cell.clear();
	  GetCell(cell,j,result[0]);
	  if (cell.empty()) {
	    index = inTable.GetColumnIndex(ColNames1[j]);
	    if (index>=0) {
	      cell.clear();
	      inTable.GetCell(cell,(unsigned int)index,i);
	      UpdateCell(cell,j,result[0]);
	    }
	  }
	}
      }
      else {
	// overwrite
        DeleteRow(result[0]);
        InsertRow(newRow,result[0]);
      }
    }
    target.clear();
    newRow.clear();
  }
  ColNames1.clear();
  ColNames2.clear();
  return NO_TABLE_ERROR;
}



int ISTable::Diff(ISTable& inTable) {
  int ret=1;
  unsigned int i;
  int errCode;
  unsigned int j;
  int index;
  vector<string> ColNames1;
  vector<string> ColNames2;
  vector<string> target;
  vector<unsigned int> sameCol1, sameCol2;
  vector<unsigned int> diffRow1, diffRow2;
  vector<unsigned int> sameRow1, sameRow2;
  string cell;
  string cell1, cell2;
  string Name1;
  string Name2;
  ofstream rpt;


  Name1=GetName();
  Name2=inTable.GetName();

  GetColumnNames(ColNames1);
  inTable.GetColumnNames(ColNames2);

  cout<<"** Compares table "<<Name1<<" and table "<<Name2<<" **"<<endl;

  cout<<"----------------- Columns report -----------------"<<endl;

  for (i=0; i<_colNames.size(); i++) {
    index = inTable.GetColumnIndex(ColNames1[i]);
    if (index<0) {
      cout.width(5);
      cout<<i<<"  < ("<<ColNames1[i]<<")"<<endl;
      ret = 0;
    }
    else {
      sameCol1.push_back(GetColumnIndex(ColNames1[i]));
      sameCol2.push_back(index);
      cout.width(5);
      cout<<i;
      cout.width(20);
      cout<<index<<"    ("<<ColNames1[i]<<")"<<endl;
    }
  }

  for (i=0; i<(inTable._colNames.size()); i++) {
    index = GetColumnIndex(ColNames2[i]);
    if (index<0) {
      cout.width(25);
      cout<<i<<"  < ("<<ColNames2[i]<<")"<<endl;
      ret = 0;
    }
  }

  cout<<endl;
  cout<<"------------------ Rows report ------------------"<<endl;
  for (i=0; i<GetNumRows(); i++) {
    // Find out what column ids are in key index
    int indexIndex = inTable.FindKeyIndex();
    for (unsigned int jj=0; jj< _listsOfColumns[indexIndex].size();
      jj++) {
      cell.clear();
      GetCell(cell,_listsOfColumns[indexIndex][jj],i);
      target.push_back(cell);
    }
    vector<unsigned int> result;
    inTable.Search(indexIndex, target,inTable._listsOfColumns[indexIndex],result, errCode);
    if (result.empty()) {
      // value for key is different
      cout.width(5);
      cout<<i<<"  <"<<endl;
      ret = 0;
    }
    else {
      // looking for other then key column for same key value
      j=0;
      cell1.clear();
      cell2.clear();
      GetCell(cell1,sameCol1[j],i);
      inTable.GetCell(cell2,sameCol2[j],result[0]);
      while(cell1 == cell2 &&j<sameCol1.size()) {
	cell1.clear();
	cell2.clear();
	GetCell(cell1,sameCol1[j],i);
	inTable.GetCell(cell2,sameCol2[j],result[0]);
	j++;
      }
      if (cell1 == cell2) {
	cout.width(5);
	cout<<i;
	cout.width(10);
	cout<<result[0]<<endl;
      }
      else {
	diffRow2.push_back(result[0]);
	cout.width(5);
	cout<<i<<"  *"<<endl;
	ret = 0;
      }
    }
    target.clear();
  }
    for (i=0; i<diffRow2.size(); i++){
    cout.width(15);
    cout<<diffRow2[i]<<"  *"<<endl;
  }
  cout<<endl;
  ColNames1.clear();
  ColNames2.clear();
  return ret;
}


void ISTable::MakeTableRectangular() {
  for(unsigned int i=0; i<_colNames.size(); i++) {
    // VLAD_DELETE_ROWS _numRows or GetNumRows()
    if (_numRows > _data[i]->size()) {
      for (unsigned int j = _data[i]->size(); j < _numRows ; j++) {
        _data[i]->push_back("");
      }
    }
  }
}


void ISTable::PrintTable()
{

    cout << endl << "Table " << _tableName <<  "  Columns: " <<
      _colNames.size() << "  Rows: " << GetNumRows() << endl << endl;
      
    cout << setw(5) << "RowNo";

    for (unsigned int i = 0; i < _colNames.size(); i++)
    {
        // VLAD HARDCODED CONST
        cout << setw(10) << _colNames[i] << " ";
    }
    cout << endl;

    for (unsigned int j = 0; j < GetNumRows(); j++)
    {
        cout << setw(5) << j;

        for (unsigned int i = 0; i < _colNames.size(); i++)
        {
            if (_data[i]->size() > j)
                // VLAD HARDCODED CONST
                cout << setw(10) << (*(_data[i]))[IntRowIndex(j)] << " ";
            else
                // VLAD HARDCODED CONST
                cout << setw(10) << " " << " ";
        }
      
        cout << endl;
    }

}


void ISTable::PrintTable(unsigned int indexNum)
{
  unsigned int i;
  tIndex::iterator pos;
  // int j;

  cout << endl << "Table " << _tableName <<  "  Columns: " << _colNames.size() << "  Rows: " << GetNumRows() << endl << endl;
      
      cout << setw(5)<< "RowNo";

  for (i = 0; i < _colNames.size(); i++)
  {
    // VLAD HARDCODED CONST
    cout << setw(10) << _colNames[i] << " ";
  }
  cout << endl;

  for (pos = _indices[indexNum].begin(); pos != _indices[indexNum].end(); ++pos)
  {
    cout << setw(5) << pos->second;

    for (i = 0; i < _colNames.size(); i++)
    {
      if (_data[i]->size() > pos->second)
        // VLAD HARDCODED CONST
	cout << setw(10) << (*(_data[i]))[IntRowIndex(pos->second)] << " ";
      else
        // VLAD HARDCODED CONST
	cout << setw(10) << " " << " ";
    }

    cout << endl;
  }

}


void ISTable::PrintTable(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex != -1)
        PrintTable(indexIndex);

}


int ISTable::WriteObject(FileNavigator * fileNav) {
  int err;
  unsigned int j;
  Word ret = 0, place = 0;
  unsigned int num;

  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;

  //  cout << "WRITEOBJECT - Writing table " << _tableName << endl;

//  err = fileNav->WriteString(_tableName, ret);

  string tableName(_tableName);
  tableName += " $$$3";

  err = fileNav->WriteString(tableName, ret);
  if (err) fileNav->PrintError(err);

  err = fileNav->WriteWord((Word) _version, place);
  if (err) fileNav->PrintError(err);

  err = fileNav->WriteWord((Word) _colNames.size(), place);
  if (err) fileNav->PrintError(err);

  // Write out the real number of rows (that excludes deleted rows)
  err = fileNav->WriteWord((Word) GetNumRows(), place);
  if (err) fileNav->PrintError(err);

  vector<string> colNamesToWrite;
  for (j = 0; j < _colNames.size(); j++) {
    colNamesToWrite.push_back(_colNames[j]);
  }
  err = fileNav->WriteStrings(colNamesToWrite, place);


  uWord * pToWrite = new uWord[_colNames.size()];
  for (j = 0; j < _colNames.size(); j++) {
    pToWrite[j] =  (uWord) _precision[j];
  }
  err = fileNav->WriteUWords((uWord *) pToWrite, (uWord) _colNames.size(), place);
  delete[] pToWrite;


  string _optsToWrite(_colNames.size(), ' ');
  for (j = 0; j < _colNames.size(); j++) {
    _optsToWrite[j] = _compare_opts[j];
  }
  err = fileNav->WriteString(_optsToWrite, place);

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    unsigned int cLen = _data[k]->size();
    unsigned int realLen = cLen;
    if (cLen == 0) {
      tempColumns.push_back("");
      err = fileNav->WriteStrings(tempColumns, place);
      tempColumns.clear();
    } else {
      for (l = 0; l < cLen; l++) {
        if (IsDelete(l))
        {
            --realLen;
        }
        else
        {
            tempColumns.push_back((*(_data[k]))[IntRowIndex(l)]);
        }
      }
      err = fileNav->WriteStrings(tempColumns, place);
      tempColumns.clear();
    }
  }

  //write indices
  num = _indexNames.size();
  err = fileNav->WriteWord((Word) num, place);
  if (err) fileNav->PrintError(err);
  vector<string> indexNameToWrite;
  if (num == 0) {
    indexNameToWrite.push_back("");
    err = fileNav->WriteStrings(indexNameToWrite, place);
  }
  else {
    for (l = 0; l < num; l++) {
      indexNameToWrite.push_back(_indexNames[l]);
    }
    err = fileNav->WriteStrings(indexNameToWrite, place);
  }
  indexNameToWrite.clear();

  uWord * uniqueToWrite = new uWord[num];
  for (j = 0; j < num; j++) {
    uniqueToWrite[j] = (uWord) _unique[j];
  }
  err = fileNav->WriteUWords((uWord *) uniqueToWrite, (uWord) num, place);
  delete[] uniqueToWrite;

  unsigned int len;
  for (l=0; l<num;l++) {
    len = _listsOfColumns[l].size();
    uWord * _listOfColumsToWrite = new uWord[len];
    for (j = 0; j < len; j++) {
      _listOfColumsToWrite[j] = (uWord) _listsOfColumns[l][j];
    }
    err = fileNav->WriteUWords((uWord *) _listOfColumsToWrite, (uWord) len, place);
    delete[] _listOfColumsToWrite;
  }

  // This version of odb file ( $$$3) and ISTables (_version == 5)
  // does not store serialized indices to the *.odb files. They will be
  // re-generated upon reading.

  colNamesToWrite.clear(); 
  return ret;
}


int ISTable::GetObject(Word index, FileNavigator * fileNav) {
/*
  There are several GetObject methods to support reading tables
  from binary files saved with some of previous verson
*/
  string tName;
  int err;
  string::size_type verStringIndex;

  Clear();
  fileNav->GetString(tName, index, err);
  index++;

  // Reverse search of a table name for a version string

  verStringIndex = tName.rfind(" $$$3");
  if (verStringIndex != string::npos)
  {
    // This version found. Remove the version string from the table name.
    tName.erase(verStringIndex);
    Rename(tName);
    return GetObjectV3(index,fileNav);
  }
  else
  {
    verStringIndex = tName.rfind(" $$$2");
    if (verStringIndex != string::npos)
    {
      // This version found. Remove the version string from the table name.
      tName.erase(verStringIndex);
      Rename(tName);
      return GetObjectV2(index,fileNav);
    }
    else
    {
      // Reverse search of a table name for a version string
      verStringIndex = tName.rfind(" $$$1");
      if (verStringIndex != string::npos)
      {
        // This version found. Remove the version string from the table name
        tName.erase(verStringIndex);
        Rename(tName);
        return GetObjectV1_1(index,fileNav);
      }
      else
      {
        Rename(tName);
        return GetObjectV1(index,fileNav);
      }
    }
  }
}


int ISTable::GetObjectV3(Word index, FileNavigator * fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;
  uWord numI = 0;
  uWord num;

  _version    = fileNav->GetWord(index, err); index++;

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  _numRows    = fileNav->GetWord(index, err); index++;
  _numDels = 0; 

  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
      _colNames.push_back(temp);
    }
  }


  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    _data.push_back(NULL);
    _data[k] = new vector<string>();
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          _data[k]->push_back(tempString[l]);
      }

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
          //free(tempColumns[l3]);
        } 
        tempColumns.clear();
      } 
      delete[] tempString; 
    }
  }

  EnlargeRowMap(_numRows);

  num = fileNav->GetWord(index, err); index++;

  vector<string> idxNamesToGet;
  fileNav->GetStrings(idxNamesToGet, index, err);
  numI = idxNamesToGet.size();
  index++;

  if (numI != num) return INTERNAL_INCONSISTENCY_ERROR;
  if (numI > 0) {
    string  temp;
    for (j = 0; j < numI; j++) {
      if (idxNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = idxNamesToGet[j];
      }
        //_indexNames.push_back(temp);
    }
  }
  if (!idxNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < numI; l2++) {
      //free(idxNamesToGet[l2]);
    }
    // if (numI > 0) idxNamesToGet.clear();
  }

  vector<uWord> uniqueToGet;
  fileNav->GetUWords(uniqueToGet, index, err);
  index++;
  if (!uniqueToGet.empty()) {
    for (j = 0; j < numI; j++) {
      _unique.push_back(uniqueToGet[j]);
    }
  }
  if (!uniqueToGet.empty()) uniqueToGet.clear();


  // This version of odb file ( $$$3) and ISTables (_version == 5)
  // does not take serialized indices from the *.odb files, but
  // re-generates them.
  vector<uWord> listToGet;
  for (j=0; j<num; j++)
  {
    fileNav->GetUWords(listToGet, index, err);
    index++;
    vector<unsigned int> list;
    for (unsigned int l2=0; l2<listToGet.size(); l2++){
      list.push_back(listToGet[l2]);
    }
    CreateIndex(idxNamesToGet[j], list);
    //_listsOfColumns.push_back(list);
    list.clear();
    if (!listToGet.empty()) listToGet.clear();
  }
  
  return ret;
}


int ISTable::GetObjectV2(Word index, FileNavigator * fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;
  uWord numI = 0;
  uWord numK = 0;
  uWord num;

  _version    = fileNav->GetWord(index, err); index++;

  if ((_version==3) || (_version==4)) {
    num = fileNav->GetWord(index, err);index++;
    int * kToGet = (int *)fileNav->GetWords(index, numK, err);index++;
    if (kToGet) delete kToGet;
  }

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  _numRows    = fileNav->GetWord(index, err); index++;
  _numDels = fileNav->GetWord(index, err); index++;

  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;

  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
      _colNames.push_back(temp);
    }
  }


  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  // VLAD: THIS CODE HAS NOT BEEN TESTED
  // NOTE: The data read here is a former ISTable member _deleted. Values
  // for _deleted used to be 0 or 1, where 0 indicated that row is not
  // deleted and 1 indicated that row was deleted.

  // Out of order read of _deleted, since it is needed for proper
  // construction of _data. Index is not advanced.
  vector<unsigned int> deleted;
  vector<uWord> dToGet;
  fileNav->GetUWords(dToGet, index + _colNames.size(), err);
  if (!dToGet.empty()) {
    for (j = 0; j < _numRows; j++) {
      deleted.push_back(dToGet[j]);
    }
  }
  if (!dToGet.empty()) dToGet.clear();

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    _data.push_back(NULL);
    _data[k] = new vector<string>();
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          if (!deleted[l])
              _data[k]->push_back(tempString[l]);
      }

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
          //free(tempColumns[l3]);
        } 
        tempColumns.clear();
      } 
      delete[] tempString; 
    }
  }

  index++;

  num = fileNav->GetWord(index, err); index++;

  vector<string> idxNamesToGet;
  fileNav->GetStrings(idxNamesToGet, index, err);
  numI = idxNamesToGet.size();
  index++;

  if (numI != num) return INTERNAL_INCONSISTENCY_ERROR;
  if (numI > 0) {
    string  temp;
    for (j = 0; j < numI; j++) {
      if (idxNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = idxNamesToGet[j];
      }
        //_indexNames.push_back(temp);
    }
  }
  if (!idxNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < numI; l2++) {
      //free(idxNamesToGet[l2]);
    }
    // if (numI > 0) free(idxNamesToGet);
  }

  vector<uWord> uniqueToGet;
  fileNav->GetUWords(uniqueToGet, index, err);
  index++;
  if (!uniqueToGet.empty()) {
    for (j = 0; j < uniqueToGet.size(); j++) {
      _unique.push_back(uniqueToGet[j]);
    }
  }
  if (!uniqueToGet.empty()) uniqueToGet.clear();

  // This version of odb file ( $$$3) and ISTables (_version == 5)
  // does not take serialized indices from the *.odb files, but
  // re-generates them.
  vector<uWord> listToGet;
  for (j=0; j<num; j++)
  {
    fileNav->GetUWords(listToGet, index, err);
    index++;
    vector<unsigned int> list;
    for (unsigned int l2=0; l2<listToGet.size(); l2++){
      list.push_back(listToGet[l2]);
    }
    CreateIndex(idxNamesToGet[j], list);
    //_listsOfColumns.push_back(list);
    list.clear();
    if (!listToGet.empty()) listToGet.clear();
  }
 
   // Correct the members
  _numRows -= _numDels;
  _numDels = 0;

  EnlargeRowMap(_numRows);

  return ret;
}


int ISTable::GetObjectV1_1(Word index, FileNavigator * fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;
  uWord numK = 0;

  _version    = fileNav->GetWord(index, err); index++;
  int * kToGet = (int *)fileNav->GetWords(index, numK, err);index++;
  if (kToGet) delete kToGet;

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  _numRows    = fileNav->GetWord(index, err); index++;
  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;
  int treeAlloc  = (int) fileNav->GetWord(index, err); index++;
  unsigned int numTrees = fileNav->GetWord(index, err); index++;
  
  EnlargeRowMap(_numRows);

  int * cMap = (int *) fileNav->GetWords(index, numC, err);index++;
  if (cMap) free(cMap);


  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
        _colNames.push_back(temp);
    }
  }
  if (!colNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < _colNames.size(); l2++) {
      //free(colNamesToGet[l2]);
    }
    if (_colNames.size() > 0) colNamesToGet.clear();
  }

  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    _data.push_back(NULL);
    _data[k] = new vector<string>();
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          _data[k]->push_back(tempString[l]);
      }

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
          //free(tempColumns[l3]);
        } 
        tempColumns.clear();
      } 
      delete[] tempString; 
    }
  }
/********************************************/
  int tableIndex = index ;
  if ((treeAlloc <= 0) || (numTrees <= 0)) {
    if (treeAlloc <= 0) treeAlloc = 0;
    if (numTrees <= 0) numTrees = 0;
    return INTERNAL_INCONSISTENCY_ERROR;
  }


  for (k = 0; k < numTrees; k++) {
    uWord numI = 0;
    int * tInts = (int *) fileNav->GetWords(tableIndex + k, numI, err); index++;

    if (numI > 0) {
      free(tInts);
    }
    tInts = NULL;
  }

  return ret;
}



int ISTable::GetObjectV1(Word index, FileNavigator * fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  _numRows    = fileNav->GetWord(index, err); index++;
  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;
  int treeAlloc  = (int) fileNav->GetWord(index, err); index++;
  unsigned int numTrees = fileNav->GetWord(index, err); index++;
  
  EnlargeRowMap(_numRows);

  int * cMap = (int *) fileNav->GetWords(index, numC, err);index++;
  if (cMap) free(cMap);


  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
        _colNames.push_back(temp);
    }
  }
  if (!colNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < _colNames.size(); l2++) {
      //free(colNamesToGet[l2]);
    }
    if (_colNames.size() > 0) colNamesToGet.clear();
  }

  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    _data[k] = new vector<string>();
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          _data[k]->push_back(tempString[l]);
      }

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
         // free(tempColumns[l3]);
        } 
        tempColumns.clear();
      } 
      delete[] tempString; 
    }
  }
/********************************************/
  int tableIndex = index ;
  if ((treeAlloc <= 0) || (numTrees <= 0)) {
    if (treeAlloc <= 0) treeAlloc = 0;
    if (numTrees <= 0) numTrees = 0;
    return INTERNAL_INCONSISTENCY_ERROR;
  }


  for (k = 0; k < numTrees; k++) {
    uWord numI = 0;
    int * tInts = (int *) fileNav->GetWords(tableIndex + k, numI, err); index++;

    if (numI > 0) {
      free(tInts);
    }
    tInts = NULL;
  }

  return ret;
}


const string ISTable::GetErrorMessage(const int errCode) {
  switch (errCode) {
  case NO_TABLE_ERROR: return "No Error";
  case ROW_OUT_OF_BOUNDS: return "WARNING: Row index out of bounds";
  case COLUMN_OUT_OF_BOUNDS: return "WARNING: Column index out of bounds";
  case NO_TREE_ON_COLUMN: return "WARNING: Column is not searchable";
  case NEW_COLUMN_LENGTH_ZERO: return "WARNING: New column has nothing";
  case ADD_UPDATE_NULL: return "WARNING: Attempting to add or update a null value to the table -- ignored";
  case COLUMN_NAME_NOT_FOUND: return "WARNING: Column name not found";
  case SOME_COLUMN_NAMES_NOT_FOUND: return "WARNING: Some column names not found -- ignoring these names";
  case REGEX_COMPILE_FAILED: return "WARNING: Regular expression compilation failed";
  case NO_APPROPRIATE_INDEX: return "WARNING: There's no appropriate index to make the search";
  case NOT_FOUND: return "WARNING: No rows are found";
  case DELETED_ROW: return "WARNING: The row is deleted";
  case INDEX_CORRUPTED: return "WARNING: The index is currupted - try to rebuild it";
  case ASSERT_WARNING: return "WARNING: Assertion on ISTable had an exception";
  case ASSERT_NULL_DATA_POINTER: return "WARNING: Assertion on ISTable showed null pointer to data -- Table may not have been initialized correctly";
  case TABLE_WARNING: return "Any value less than this is a warning"; 
  case NULL_COMPARISON: return "ERROR: Null comparison made -- cannot continue comparison";
  case DOUBLE_CONVERSION_ERROR: return "ERROR: Could not convert string to double";
  case INTEGER_CONVERSION_ERROR: return "ERROR: Could not convert string to double"; 
  case NULL_SEARCH_LIST: return "ERROR: Cannot search a list that is null"; 
  case NOT_A_DATATYPE_ERROR: return "ERROR: Attempting to change column to OR column has an illegal datatype";
  case ERROR_NO_FILE_NAVIGATOR: return "ERROR: Cannot write or get an object without an instance of FileNavigator";
  case INTERNAL_INCONSISTENCY_ERROR: return "ERROR: There is an internal class error -- may lead to unpedictable behavior";
  default: return "Unknown error code";

  }

}


void ISTable::ConvertToInt(const string& a, string& ret)
{

  ret.clear();

  const unsigned int len = a.size();

  if (len > INT_LIMIT)
  {
    return;
  }

  int intValue = 0;

  if (len != 0)
  {
      intValue = String::StringToInt(a);
  }

  if (intValue < 0)
  {
      const string maxNum(len - 1, '9');

      ret.push_back('-');
      ret.append(INT_LIMIT - len, '9');

      // maxNum - (-intValue)
      ret += String::IntToString(String::StringToInt(maxNum) + intValue);
  }
  else
  {
      ret.assign(INT_LIMIT - len, '0');
      ret += a;
  }

}


void ISTable::ConvertDouble(const string& a, string& ret){
/* This method convert string representing a double value into 
   another string. when we wont to compare two double value
   we can compare two string value (converted value) and
   result will be the same
*/

  const int MAX_DECIMAL_DIGIT = 9;
  string m;  // Mantissa before decimal point
  string r;  // Mantissa after decimal point (remainder)
  string e;  // Exponent, without letter "e"
  string aa; // Exponential representation of the input double
  // Maximum remainder complement value

  int eint;  // Exponent value
  double adouble; // Input value


  // Convert the original double representation to mantissa/exponent
  // representation.
  adouble = String::StringToDouble(a);

  ostringstream outStringStream;
  outStringStream << std::scientific << std::showpoint <<
    std::setprecision(MAX_PRECISION) << adouble;
    
  aa = outStringStream.str();

  // Extract mantissa, remainder and exponent
  string::size_type dotCharIndex = aa.find('.');
  if (dotCharIndex == string::npos)
  {
    return;
  }

  m.append(aa, 0, dotCharIndex);

  string::size_type expCharIndex = aa.find('e');
  if (expCharIndex == string::npos)
  {
    // VLAD BUG: Fix this
    return;
  }

  r.append(aa, dotCharIndex + 1, expCharIndex - dotCharIndex - 1);
  e.append(aa, expCharIndex + 1, aa.size() - expCharIndex - 1);
 
  eint = String::StringToInt(e);

  if (adouble < 0)
  {
    ret.push_back('-');
  }
  else
  {
    ret.push_back('0');
  }

  if (eint < 0)
  {
    const string maxExpNum(e.size() - 1, '9');

    ret.push_back('-');

    // Complement the exponent
    ret.append(EXPONENT - e.size(), '9');

    // maxExpNum - (-eint)
    ret += String::IntToString(String::StringToInt(maxExpNum) + eint);
  }
  else
  {
    ret.push_back('0');
    ret.append(EXPONENT - e.size(), '0');
    ret.append(e, 1, e.size() - 1);
  }

  if (adouble < 0)
  {
    // Complement mantissa
    // MAX_DECIMAL_DIGIT - (-mantissa before dot)
    ret += String::IntToString(MAX_DECIMAL_DIGIT + String::StringToInt(m));

    ret.push_back('.');

    // Complement each digit in the remainer
    for (unsigned int index = 0; index < r.size(); ++index)
    {
      ret += String::IntToString(MAX_DECIMAL_DIGIT -
        String::StringToInt(r.substr(index, 1)));
    }
  }
  else
  {
    ret += m;
    ret += '.';
    ret += r;
  }

}


void ISTable::ConvertToLowerNoWhiteSpace(const string& a, string& ret){
  string temp;
  String::RemoveWhiteSpace(a, temp);
  String::LowerCase(temp, ret);
}


void ISTable::EnlargeRowMap(const unsigned int numRows)
{

    unsigned int firstIndex = _rowMap.size();

    for (unsigned int rowIndex = firstIndex;
      rowIndex < (firstIndex + numRows); rowIndex++)
    {
        _rowMap.push_back(rowIndex);
    }

}


void ISTable::ReduceRowMap(const unsigned int numRows)
{

    _rowMap.erase(_rowMap.begin() + numRows, _rowMap.end());

}


void ISTable::MarkRowDeleted(const unsigned int rowIndex)
{

    unsigned int saveRowIndex = _rowMap[rowIndex];

    unsigned int index;

    for (index = rowIndex; index < (GetNumRows() - 1); index++)
    {
        _rowMap[index] = _rowMap[index + 1];
    }

    _rowMap[index] = saveRowIndex;

    ++_numDels;

}


void ISTable::UnMarkRowDeleted(const unsigned int rowIndex)
{

    unsigned int availRowIndex = _rowMap[GetNumRows()];

    unsigned int index;

    for (index = GetNumRows(); index > rowIndex; --index)
    {
        _rowMap[index] = _rowMap[index - 1];
    }

    _rowMap[index] = availRowIndex;

    --_numDels;

}


bool ISTable::IsDelete(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        return(true);
    }
    else
    {
        return(false);
    }

}


StringCompare::eCompareType 
  ISTable::GetCompareType(const vector<unsigned int>& colIndices)
{

    // Initially set it to case sensitive type.
    StringCompare::eCompareType compareType = StringCompare::eCASE_SENSITIVE;

    // In setting the compare type, only the options of the column
    // at index 0 are considered. It is assumed that all the rest of
    // the columns have the same options as the first column.

    unsigned char opts = _compare_opts[colIndices[0]];
    switch ((opts & DT_MASK) >> 4)
    {
        case DT_STRING_VAL:
            if (opts & SC_MASK)
            {
                // Case in-sensitive column.
                compareType = StringCompare::eCASE_INSENSITIVE;
            }
            break;
        default:
            break;
    }

    return(compareType);

}


void ISTable::UpdateColListOnColInsert(const unsigned int colIndex)
{

    // Updating lists of columns. If a column in the list has column
    // index >= colIndex then increment it by 1
    for (unsigned int i = 0; i < _listsOfColumns.size(); i++)
    {
        for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
        {
            if (_listsOfColumns[i][j] >= colIndex)
                _listsOfColumns[i][j]++;
        }
    }

}


void ISTable::UpdateColListOnColDelete(const unsigned int colIndex)
{

    // Updating lists of columns. If a column in the list has column
    // index >= colIndex then decrement it by 1
    for (unsigned int i = 0; i < _listsOfColumns.size(); i++)
    {
        for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
        {
            if (_listsOfColumns[i][j] >= colIndex)
                _listsOfColumns[i][j]--;
        }
    }

}


void ISTable::UpdateColListOnRowAppend(const unsigned int rowIndex, 
  const unsigned int fromColIndex, const unsigned int toColIndex)
{

    // If theElement is put in column (i) wich is a part of one or more
    // indices, then those indices have to be updated
    for (unsigned int j = 0; j < _indices.size(); j++)
    {
        for (unsigned int k = 0; k < _listsOfColumns[j].size(); k++)
        {
            if ((_listsOfColumns[j][k] >= fromColIndex) &&
              (_listsOfColumns[j][k] <= toColIndex))
            {
                // VLAD: EXAMINE IF THIS IS NEEDED IN EVERY ITERATION
    	        UpdateIndex(j, rowIndex);
                break;
            }
        }
    }

}


void ISTable::UpdateColListOnCellUpdate(const unsigned int rowIndex,
  const unsigned int colIndex)
{

    for (unsigned int j = 0; j < _indices.size(); j++)
    {
        for (unsigned int k = 0; k < _listsOfColumns[j].size(); k++)
        {
            if (_listsOfColumns[j][k] == colIndex)
            {
                UpdateIndex(j, rowIndex);
                // VLAD: EXAMINE THIS !!
                break;
            }
        }
    }

}

