/*
FILE:     TableFile.C
*/
/*
VERSION:  3.001
*/
/*
DATE:     5/18/2006
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2006 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
/*
  PURPOSE:    Class for tables management in persistent store
*/


#include <string>
#include <iostream>
#include "GenString.h"
#include "TableFile.h"


TableFile::Table::Table(const StringCompare::eCompareType caseSense)
{
  _block.clear();
  _name.clear();
  _id.clear(); 
  _indexInFile = 0;  
  _indexInMemory = 0;
  _numUsedIndicesInFile = 0;  
  _modified = 0;  
  _caseSense = caseSense;
}


TableFile::Table::Table(const string& c,
  const StringCompare::eCompareType caseSense)
{
  
  _block.clear();
  _name.clear();
  _id = c;
  _indexInFile = 0;  
  _indexInMemory = 0;
  _numUsedIndicesInFile = 0;  
  _modified = 0;  
  _caseSense = caseSense;
}


TableFile::Table::Table(const Table& t){
  
  _block = t._block;
  _name = t._name;
  _id = t._id;
  _indexInFile = t._indexInFile;  
  _indexInMemory = t._indexInMemory;     
  _numUsedIndicesInFile = t._numUsedIndicesInFile;
  _modified = t._modified;  
  _caseSense = t._caseSense;
}


TableFile::Table::~Table() {
  Clear();
}


void TableFile::Table::DeleteElement()
{
    Clear();
}


void TableFile::Table::Print()
{
  cout<<_id<<endl;
}


void TableFile::Table::Clear() {
  _block.clear();
  _name.clear();
  _id.clear();
}


void TableFile::Table::operator=(const Table& t) {
  Clear();
  _block = t._block;
  _name = t._name;
  _id = t._id;
  _indexInFile = t._indexInFile;  
  _indexInMemory = t._indexInMemory;     
  _numUsedIndicesInFile = t._numUsedIndicesInFile; 
  _modified = t._modified;  
  _caseSense = t._caseSense;
}


bool TableFile::Table::operator<(const Table& t) const
{

    return(String::IsLess(_id, t._id, _caseSense));

}


bool TableFile::Table::operator==(const Table &t)
{

    return(String::IsEqual(_id, t._id, _caseSense));

}


TableFile::Block::Block(const StringCompare::eCompareType caseSense)
{
  _name.clear();
  _caseSense = caseSense;
}


TableFile::Block::Block(const string& c,
  const StringCompare::eCompareType caseSense)
{
  _name = c;
  _caseSense = caseSense;
}

TableFile::Block::Block(const Block& b){
  
  _name = b._name;
  _caseSense = b._caseSense;
}


TableFile::Block::~Block() {
  Clear();
}


void TableFile::Block::DeleteElement() {
  Clear();
}


void TableFile::Block::Print() {
  cout<<_name<<endl;
}


void TableFile::Block::Clear() {
  _name.clear();
}


void TableFile::Block::operator=(const Block& b) {
  Clear();
  _name = b._name;
  _caseSense = b._caseSense;
}


bool TableFile::Block::operator<(const Block& b) const
{

    return(String::IsLess(_name, b._name, _caseSense));

}


bool TableFile::Block::operator==(const Block &b)
{

    return(String::IsEqual(_name, b._name, _caseSense));

}


TableFile::TableFile(const StringCompare::eCompareType caseSense)
{
    Clear();
    _caseSense = caseSense;
}


TableFile::TableFile(const string& fileName, const eFileMode fileMode,
  const StringCompare::eCompareType caseSense)
{
  Clear();
  _caseSense = caseSense;
  OpenFile(fileName, fileMode);

}


TableFile::~TableFile()
{
      CloseFile(false);
      Reset();
}


void TableFile::Clear() {
  _f = NULL;
  _currBlockName.clear();
  _fileMode = NO_MODE;
  _tableObjs.clear();
  _statusInd = eCLEAR_STATUS;
  SetCurrentBlock(" # 1");
}


void TableFile::Reset() {
  unsigned int i,j;
  if (_f) delete _f;
  _fileName.clear();
  _currBlockName.clear();

  if (!_tableObjs.empty()) {
    for (i = 0, j=0; j < _tables.size(); i++) {  
	if (_tables[i].GetIndexInMemory() > -1) {
	  if (_tableObjs[_tables[i].GetIndexInMemory()]) { 
	    delete _tableObjs[_tables[i].GetIndexInMemory()];
	    _tableObjs[_tables[i].GetIndexInMemory()] = NULL;
	  }
	}
	j++;
    }
    _tableObjs.clear(); 
  }
  Clear();
}


int TableFile::OpenFile(const string& fileName, const eFileMode fileMode) {
  int ret;

  _fileMode = fileMode;

  _fileName = fileName;

  if (_fileMode != WRITE_MODE) {
  FileNavigator * fnav = new FileNavigator();
    if (((ret = fnav->OpenFile(fileName, fileMode,0)) < 0) || ((ret = fnav->ReadFileHeader()) < 0)) {
      delete fnav;
      fnav = NULL;
      return ret;
    }

    if (!_f) {
      _f = fnav;
      fnav = NULL;
    } else {
      _f->CloseFile();
      delete _f;
      _f = fnav;
      fnav = NULL;
    }
    return _ReadFileIndex();
    
  } else
    return 0;
}


void TableFile::CloseFile(const bool serialize)
{

    if (_fileMode == READ_MODE)
    {
        return;
    }

    if ((_fileMode == WRITE_MODE) && serialize)
    {
        FileNavigator* fnav = new FileNavigator();

        fnav->OpenFile(_fileName, _fileMode, 0);
        fnav->ReadFileHeader();

        if (_f == NULL)
        {
            _f = fnav;
        }
        else
        {
            _f->CloseFile();
            delete _f;
            _f = fnav;
        }
    }

    if (_f == NULL)
    {
        return;
    }

    int loc;

    // Write all modified tables 
    for (unsigned int tableI = 0; tableI < _tables.size(); tableI++)
    {
        if (_tables[tableI].GetModified())
        {
            ISTable* tableP = _tableObjs[_tables[tableI].GetIndexInMemory()];
            loc = tableP->WriteObject(_f);
            _tables[tableI].SetIndexInFile(loc);
            // ERROR HANDLING
            // if (loc < 0)
            //    return -344;
        }
    }

    _WriteFileIndex();

    _f->CloseFile();
    delete _f;

    _f = NULL;

    _fileMode = NO_MODE;

}


bool TableFile::IsBlockPresent(const string& blockName)
{

    if (blockName.empty())
        return(false);

    if (_blocks.empty())
        return(false);

    Block block(blockName, _caseSense);

    vector<Block>::iterator pos = find(_blocks.begin(),
      _blocks.end(), block);
    if (pos != _blocks.end())
    {
        // Found it.
        return(true);
    }
    else
    {
        // Not found.
        return(false);
    }

} // End of TableFile::IsBlockPresent()


string TableFile::AddBlock(const string& blockName)
{

    if (blockName.empty())
    {
        // Set the flag that empty block name is found
        _statusInd |= eUNNAMED_BLOCKS;
    }

    string addedBlockName = blockName;

    if (blockName.empty() || IsBlockPresent(blockName))
    {
        // This is either a nameless block or a block that is already present.

        // Set the flag that duplicate blocks are found
        _statusInd |= eDUPLICATE_BLOCKS;

        // Create a new block name.
        addedBlockName += " # ";
        addedBlockName += String::IntToString(_blocks.size() + 1);
    }

    // Add the new block
    _AddBlock(addedBlockName);

    return(addedBlockName);

} // End of TableFile::AddBlock()


void TableFile::SetCurrentBlock(const string& blockName)
{

    // VLAD ERROR HANDLING EXCEPTION (if block is empty)
    if (blockName.empty())
        return;

    _currBlockName = blockName;

} // End of TableFile::SetCurrentBlock


string TableFile::GetFirstBlockName()
{

    string ret;

    if (_blocks.empty())
        return ret;

    return(_blocks[0].GetName());

} // End of TableFile::GetFirstBlockName()


void TableFile::GetBlockNames(vector<string>& blockNames)
{

    blockNames.clear();

    if (_blocks.empty())
        return;

    vector<Block>::iterator pos;

    for (pos = _blocks.begin(); pos != _blocks.end();
      ++pos)
    {
        blockNames.push_back(pos->GetName());
    }

}


int TableFile::RenameBlock(const string& oldBlockName,
  const string& newBlockName)
{

    int ret = -1;

    if (_blocks.empty())
    {
        return(ret);
    }

    int blockIndex = GetBlockIndex(oldBlockName);
    if (blockIndex < 0)
    {
        return(ret);
    }

    // Rename block names in all present tables
    for (unsigned int tableI = 0; tableI < _tables.size(); ++tableI)
    {
        if (_tables[tableI].GetBlock() == oldBlockName)
        {
            _tables[tableI].SetBlock(newBlockName); 
        }
    }

    Block block(newBlockName, _caseSense);

    _blocks[blockIndex] = block;

    return(1);

} // End of TableFile::RenameBlock()


int TableFile::WriteTable(ISTable* s, const string& blockName,
  const string& tableName)
{

    if (_fileMode == READ_MODE)
        return(-177);

    if ((s == NULL) || tableName.empty())
        return(-1);

    _AddBlock(blockName);

    SetCurrentBlock(blockName);
   
    string tableId = _BuildTableId(blockName, tableName);
    if (tableId.empty())
        return(-1);

    int tableIndex = _GetTableIndex(tableId);

    if (tableIndex >= 0)
    {
        _tables[tableIndex].SetModified(true);
        if (_tableObjs[_tables[tableIndex].GetIndexInMemory()] != s)
            delete _tableObjs[_tables[tableIndex].GetIndexInMemory()];
        _tableObjs[_tables[tableIndex].GetIndexInMemory()] = s;
    }
    else 
    {
        int newIndexInMemory = _tableObjs.size();

        _tableObjs.push_back(NULL);
        _tableObjs[newIndexInMemory] = s;

        // Add new table to file index
        Table table(_caseSense);

        table.SetBlock(blockName);
        table.SetName(tableName);
        table.SetId(tableId);

        table.SetNumUsedIndicesInFile(12 + s->GetNumColumns() +
          s->GetNumIndices() * (5 + s->GetNumRows() * 6));

        table.SetIndexInMemory(newIndexInMemory);

        table.SetModified(true);

        _tables.push_back(table);
    }

    return(1);

}


bool TableFile::IsTablePresent(const string& blockName,
  const string& tableName)
{

  string tableId;

  if (_tables.empty()) return false;
  tableId = _BuildTableId(blockName, tableName);
  if (tableId.empty()) return false;

  Table a(tableId, _caseSense);
  vector<Table>::iterator pos = find(_tables.begin(),
    _tables.end(), a);
  if (pos != _tables.end())
  {
      // Found it.
      return true;
  }
  else
  {
      // Not found.
      return false;
  }

}


ISTable* TableFile::GetTablePtr(const string& blockName,
  const string& tableName)
{

    string tableId = _BuildTableId(blockName, tableName);
    if (tableId.empty())
    {
        return NULL;
    }

    int tableIndex = _GetTableIndex(tableId);

    if (tableIndex < 0)
        return NULL;

    int indexInMemory = _tables[tableIndex].GetIndexInMemory();

    int lastTableIndexInMemory = _tableObjs.size() - 1;

#if DEBUG
    cerr << "GetTablePtr indexInMemory " << indexInMemory <<
      " lastTableIndexInMemory " << lastTableIndexInMemory << endl;
#endif

    if (indexInMemory >= 0)
    {
        return _tableObjs[indexInMemory];
    }
    else
    {
        int newTableIndexInMemory = lastTableIndexInMemory + 1;

        _tables[tableIndex].SetIndexInMemory(newTableIndexInMemory);

        _tableObjs.push_back(NULL);
        _tableObjs[newTableIndexInMemory] = new ISTable();
        _tableObjs[newTableIndexInMemory]->\
          GetObject(_tables[tableIndex].GetIndexInFile(), _f);

        // Number of file navigator logical index slots per table ...
        _tables[tableIndex].SetNumUsedIndicesInFile(12 + 
          _tableObjs[newTableIndexInMemory]->GetNumColumns() + 
          _tableObjs[newTableIndexInMemory]->GetNumIndices() *
          (5 + _tableObjs[newTableIndexInMemory]->GetNumRows() * 6));

        return(_tableObjs[newTableIndexInMemory]);
    }

}


int TableFile::DeleteTable(const string& blockName, const string& tableName)
{

    string tableId = _BuildTableId(blockName, tableName);
    if (tableId.empty())
        return(TABLE_NOT_FOUND);

    int tableIndex = _GetTableIndex(tableId);
    if (tableIndex < 0)
        return(-1);

    if (_fileMode == UPDATE_MODE)
    {
        // Delete the table from persistent store ..
        int indexInFile = _tables[tableIndex].GetIndexInFile();
        for (int j = 0;
          j < _tables[tableIndex].GetNumUsedIndicesInFile(); j++)
        {
            _f->Delete(indexInFile + j);
        }
    }

    int indexInMemory = _tables[tableIndex].GetIndexInMemory();

    if (_tableObjs[indexInMemory] != NULL)
        delete _tableObjs[indexInMemory];

    _tableObjs[indexInMemory] = NULL;

    _tableObjs.erase(_tableObjs.begin() + indexInMemory);

    for (unsigned int i = 0; i < _tables.size(); i++)
    {
        if (_tables[i].GetIndexInMemory() > indexInMemory)
            _tables[i].SetIndexInMemory(_tables[i].GetIndexInMemory() - 1);
    }

    Table table(_caseSense);

    table.SetId(tableId);

    vector<Table>::iterator pos = find(_tables.begin(), _tables.end(), table);
    if (pos != _tables.end())
    {
        _tables.erase(pos);
    }

    return(0);

}


void TableFile::GetTableNames(vector<string>& tableNames)
{

    tableNames.clear();

    if (_tables.empty())
        return;

    for (unsigned int tableI = 0; tableI < _tables.size(); tableI++)
    {
        tableNames.push_back(_tables[tableI].GetName());
    }

}


void TableFile::GetTableNames(vector<string>& tableNames,
  const string& blockName)
{

    tableNames.clear();

    if (blockName.empty())
        return;

    if (_tables.empty())
        return;

    string tableName;
    string localBlockName;

    for (unsigned int tableI = 0; tableI < _tables.size(); tableI++)
    {
        tableName = _tables[tableI].GetName();
        localBlockName = _tables[tableI].GetBlock();

        if (tableName.empty() || localBlockName.empty())
        {
            continue;
        }

        if (blockName == localBlockName)
        {
  	    tableNames.push_back(tableName);
        }
    }

}


void TableFile::PrintHeaderInfo()
{

    cout << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"\
      "++++++++++++++++++++++++++" << endl;
    cout << "TableFile::PrintHeaderInfo*TableFile::PrintHeaderInfo*"\
      "TableFile::PrintHeaderInfo" << endl;
    cout << "File name: " << _fileName << endl;
    cout << "File mode: " << _fileMode << endl;
    cout << "Current Data Block: " << _currBlockName << endl;
    cout << "numBlocks = " << _blocks.size() << endl;
    cout << "numTables = " << _tables.size() << endl;

    for (unsigned int i = 0; i < _blocks.size(); i++)
    {
        cout << "++++++++++++++++++++++++++++++++++++++++++++++++++++++"\
          "+++++++++++++++++" << endl;
        cout << "Block [" << i << "] " << GetBlockName(i) << endl; 
    }

    for (unsigned int i = 0; i < _tables.size(); i++)
    {
        cout << "++++++++++++++++++++++++++++++++++++++++++++++++++++++"\
          "+++++++++++++++++" << endl;
        cout << "Table [" << i << "] " << _tables[i].GetId() << endl; 
        cout << "block = " << _tables[i].GetBlock() << endl; 
        cout << "name = " << _tables[i].GetName() << endl; 
        cout << "indexInFile = " << _tables[i].GetIndexInFile() << endl; 
        cout << "indexInMemory = " << _tables[i].GetIndexInMemory() << endl; 
        cout << "numUsedIndicesInFile = " << _tables[i].GetNumUsedIndicesInFile() << endl;
    }

    cout << "TableFile::PrintHeaderInfo*TableFile::PrintHeaderInfo*"\
      "TableFile::PrintHeaderInfo" << endl;
    cout << "eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+eod+"\
      "eod+eod+eod+eod+eod+eod" << endl;

}






void TableFile::_AddBlock(const string& blockName)
{

    // VLAD ERROR HANDLING EXCEPTION (if block is already present)
    if (!IsBlockPresent(blockName))
    {
        Block bi(_caseSense);

        bi.SetName(blockName);

        _blocks.push_back(bi);
    }

} // End of TableFile::_AddBlock()


int TableFile::GetBlockIndex(const string& blockName)
{

    int ret = -1;

    if (_blocks.empty())
        return -1;

    Block a(blockName, _caseSense);

    vector<Block>::iterator pos = find(_blocks.begin(),
      _blocks.end(), a);
    if (pos != _blocks.end())
    {
        ret = pos - _blocks.begin();
    }

    return ret;

}


string TableFile::GetBlockName(int blockIndex)
{

    string ret;

    if (_blocks.empty() ||
      (blockIndex > (int)(_blocks.size())))
        return ret;

    return(_blocks[blockIndex].GetName());

}


int TableFile::_GetTableIndex(const string& tableId)
{  // search the internal identifier

    if (tableId.empty())
        return -100;
    if (_tables.empty())
        return -1;

    int ret = -1;

    Table a(tableId, _caseSense);
    vector<Table>::iterator pos = find(_tables.begin(),
      _tables.end(), a);
    if (pos != _tables.end())
    {
        ret = pos - _tables.begin();
    }

    return ret;

}


string TableFile::_GetBlockId(const string& blockName)
{
    // Block id is the string representation of block index.

    string blockId;

    if (blockName.empty())
        return(blockId);

    int blockIndex = GetBlockIndex(blockName);
    if (blockIndex >= 0)
    {
        blockId = String::IntToString(blockIndex);
    }

    return(blockId);

} // End of TableFile::_GetBlockId()



string TableFile::_GetBlockName(const string& blockId) {
  int blockIndex;
 
  if (blockId.empty()) return "";
  blockIndex = String::StringToInt(blockId);
  return (GetBlockName(blockIndex));
}


string TableFile::_BuildTableId(const string& blockName,
  const string& tableName)
{

    string tableId;

    if (tableName.empty())
        return(tableId); 

    if (blockName.empty())
    {
        // Handle null block gracefully.
        tableId = "__";
        tableId += tableName;
    }
    else
    {
        string blockId = _GetBlockId(blockName);
        if (!blockId.empty())
        {
            tableId = "_";
            tableId += blockId;
            tableId += "_";
            tableId += tableName;
        }
    }

    return(tableId);

}


void TableFile::_GetTableIds(vector<string>& tableIds) {
  tableIds.clear();
  if (_tables.empty()) return;

  unsigned int i = 0, j=0;
  while(j < _tables.size()) {
      tableIds.push_back(_tables[i].GetId());
      j++;
    i++;
  }
}
void TableFile::_GetTableBlockNames(vector<string>& tableBlockNames) {
  tableBlockNames.clear();
  if (_tables.empty()) return;

  unsigned int i = 0, j = 0;
  while(j < _tables.size()) {
      tableBlockNames.push_back(_tables[i].GetBlock());
      j++;
    i++;
  }
}


int TableFile::_WriteFileIndex()
{

    unsigned int i, j;
    int where = 0, ret = 0, err=0;
    vector<string> blockNames;
    vector<string> tableNames;

    if (!_f)
        return -1;
    if (_fileMode == READ_MODE)
        return -1;

    int* tableLocs = new int[_tables.size()];
    int* bOrder = new int[1];
    int* tOrder = new int[1];

    for (i = 0, j = 0; j < _tables.size(); i++)
    {
        tableLocs[j] = _tables[i].GetIndexInFile();
        j++;
    }

    vector<string> tableIds;
    _GetTableIds(tableIds);
    GetTableNames(tableNames);
    vector<string> tableBlockNames;
    _GetTableBlockNames(tableBlockNames);
    GetBlockNames(blockNames);
  
    // Append the file index and clean up the in-memory table cache and index.

    err = _f->WriteStrings(tableIds, (Word &) where);
    err += _f->WriteStrings(tableNames,  (Word &) ret);
    err += _f->WriteStrings(tableBlockNames,  (Word &) ret);
    err += _f->WriteStrings(blockNames,  (Word &) ret);
    err += _f->WriteWords((Word *) tOrder, 0, (Word &) ret);
    err += _f->WriteWords((Word *) bOrder, 0, (Word &) ret);
    err += _f->WriteWords((Word *) tableLocs, _tables.size(), (Word &) ret);

    // Clean-up  

    if (tableLocs) delete[] tableLocs;
    if (bOrder)    delete[] bOrder;
    if (tOrder)    delete[] tOrder;

    tableIds.clear();
    tableNames.clear();
    tableBlockNames.clear();
    blockNames.clear();

    return err;

}


int TableFile::_ReadFileIndex()
{
    int err = 0, numTables = 0, numBlocks = 0, where = -1;
    int bOrderLen, tOrderLen;
    unsigned int i;
    string::size_type hashSymbolIndex;

    if (!_f)
        return FILE_NOT_OPEN;
    if ((_fileMode != READ_MODE) && (_fileMode != UPDATE_MODE))
        return FILE_NOT_READ_MODE;

    where = _f->Length() - 7; 
    if (where <  0 )
    {
        return FILE_HEADER_INCONSISTENCY;
    }

    vector<string> tableIds;
    _f->GetStrings(tableIds, where, err);
    numTables = tableIds.size();
    if (err < 0) return err;

    vector<string> tableNames;
    _f->GetStrings(tableNames, where + 1, err);
    numTables = tableNames.size();
    if (err < 0) return err;

    vector<string> tableBlockNames;
    _f->GetStrings(tableBlockNames, where + 2, err);
    numTables = tableBlockNames.size();
    if (err < 0) return err;

    vector<string> blockNames;
    _f->GetStrings(blockNames, where + 3, err);
    numBlocks = blockNames.size();
    if (err < 0) return err;

    // Set the status indicator based on the read information
    for (i = 0; i < (uWord)numBlocks; i++)
    {
        hashSymbolIndex = blockNames[i].find('#');
        if (hashSymbolIndex != string::npos)
        {
            // Found a block with a hash symbol in its name
            if (hashSymbolIndex > 1)
            {
                // Set the flag that duplicate blocks are found
                _statusInd |= eDUPLICATE_BLOCKS;
            } // More than a space symbol prior to a hash 
            else
            {
                // Set the flag that empty block name is found
                _statusInd |= eUNNAMED_BLOCKS;
            } // Only a space symbol prior to a hash
        }
    } // For all blocks

    SetCurrentBlock(blockNames[0]);

    int* tOrder2 = (int*) _f->GetWords(where + 4, (uWord &) tOrderLen, err);
    if (err < 0) return err;

    int* bOrder2 = (int*) _f->GetWords(where + 5, (uWord &) bOrderLen, err);
    if (err < 0) return err;

    int* locs = (int*) _f->GetWords(where + 6, (uWord &) numTables, err);
    if (err < 0) return err;

    if (tOrderLen == 0)
    {
        for (i = 0; i < (unsigned int)numTables; i++)
        {
            Table ti(_caseSense);

            ti.SetId(tableIds[i]);
            ti.SetName(tableNames[i]);
            ti.SetBlock(tableBlockNames[i]);
            ti.SetIndexInFile(locs[i]);
            ti.SetIndexInMemory(-1);
            ti.SetNumUsedIndicesInFile(-1);
            ti.SetModified(false);

            _tables.push_back(ti);
        }
    }
    else
    {
        int* tOrder = new int[numTables];
	for (i = 0; i < (unsigned int)numTables; i++)
        {
            tOrder[tOrder2[i]]=i;
	}
	for (i = 0; i < (unsigned int)numTables; i++)
        {
            Table ti(_caseSense);

            ti.SetId(tableIds[tOrder[i]]);
            ti.SetName(tableNames[tOrder[i]]);
            ti.SetBlock(tableBlockNames[tOrder[i]]);
            ti.SetIndexInFile(locs[tOrder[i]]);
            ti.SetIndexInMemory(-1);
            ti.SetNumUsedIndicesInFile(-1);
            ti.SetModified(false);

            _tables.push_back(ti);
        }
	if (tOrder) delete[] tOrder; 
	tOrder = NULL;
    }

    if (bOrderLen == 0)
    {
        for (i = 0; i < (unsigned int) numBlocks; i++)
        {
            Block bi(_caseSense);
            bi.SetName(blockNames[i]);
            _blocks.push_back(bi);
        }
    }
    else
    {
        int* bOrder = new int[_blocks.size()];
        for (i = 0; i < _blocks.size(); i++)
        {
            bOrder[bOrder2[i]]=i;
        }
        for (i = 0; i < _blocks.size(); i++)
        {
            Block bi(_caseSense);
            bi.SetName(blockNames[bOrder[i]]);
            _blocks.push_back(bi);
        }
        if (bOrder)
            delete[] bOrder; 
        bOrder = NULL;
    }

    // Clean-up ... 

    if (locs) free(locs); 
    locs = NULL;

    if (tOrder2) free(tOrder2); 
    tOrder2 = NULL;

    if (bOrder2) free(bOrder2); 
    bOrder2 = NULL;

    tableIds.clear(); 
    tableBlockNames.clear(); 
    tableNames.clear();
    blockNames.clear();

    return 0;

}

